import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.replace import Replace
from koheesio.spark.utils import show_string

pytestmark = pytest.mark.spark


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # case 0 : no target_column -> should replace the original column
            dict(column="input_column", from_value="hello", to_value="programmer"),
            # expected output
            [
                dict(id=1, input_column="programmer"),
                dict(id=2, input_column="world"),
                dict(id=3, input_column=None),
            ],
        ),
        (
            # case 1 : with target_column -> should add a new column
            dict(column="input_column", from_value="hello", to_value="programmer", target_column="output"),
            # expected output
            [
                dict(id=1, input_column="hello", output="programmer"),
                dict(id=2, input_column="world", output="world"),
                dict(id=3, input_column=None, output=None),
            ],
        ),
        (
            # case 2 : using aliases for input
            {"column": "input_column", "from": "hello", "to": "programmer", "target_column": "output"},
            # expected output
            [
                dict(id=1, input_column="hello", output="programmer"),
                dict(id=2, input_column="world", output="world"),
                dict(id=3, input_column=None, output=None),
            ],
        ),
        (
            # case 3 : no from_value should replace the Null / None
            dict(column="input_column", to_value="NotNone", target_column="output"),
            # expected output
            [
                dict(id=1, input_column="hello", output="hello"),
                dict(id=2, input_column="world", output="world"),
                dict(id=3, input_column=None, output="NotNone"),
            ],
        ),
        (
            # case 4 : run against non-string column
            dict(column="id", from_value="1", to_value="Numeric replace"),
            # expected output
            [
                dict(id="Numeric replace", input_column="hello"),
                dict(id="2", input_column="world"),
                dict(id="3", input_column=None),
            ],
        ),
    ],
)
def test_base(input_values, expected, spark):
    input_df = spark.createDataFrame([[1, "hello"], [2, "world"], [3, None]], ["id", "input_column"])
    df = Replace(**input_values).transform(input_df)

    actual = [k.asDict() for k in df.collect()]
    assert actual == expected


@pytest.mark.parametrize(
    "input_values",
    [
        dict(column="byte", from_value="1"),
        dict(column="short", from_value="1"),
        dict(column="integer", from_value="1"),
        dict(column="long", from_value="1"),
        dict(column="float", from_value="1.0"),
        dict(column="double", from_value="1.0"),
        dict(column="decimal", from_value="1.0"),
        dict(column="binary", from_value=b"a", happy_flow=False),
        dict(column="boolean", from_value="1", happy_flow=False),
        dict(column="timestamp", from_value="2023-01-01 00:01:01"),
        dict(column="date", from_value="2023-01-01"),
        dict(column="array", from_value="1", happy_flow=False),
        dict(column="map", from_value="1", happy_flow=False),
        dict(column="void", from_value="1", happy_flow=False),
        dict(column="string", from_value="a"),
    ],
)
def test_all_data_types(input_values, df_with_all_types):
    log = LoggingFactory.get_logger("test_all_data_types")
    happy_flow = input_values.pop("happy_flow", True)
    column = input_values["column"]

    if happy_flow:
        input_values["to_value"] = input_values.get("to_value", "happy")
        expected = input_values["to_value"]
        df = Replace(**input_values).transform(df_with_all_types)
        log.info(f"show df: \n{show_string(df, 20, 20, False)}")
        actual = df.head().asDict()[column]
        assert actual == expected
    else:
        input_values["to_value"] = "unhappy"
        expected = df_with_all_types.head().asDict()[column]  # stay the same
        df = Replace(**input_values).transform(df_with_all_types)
        log.info(f"show df: \n{show_string(df, 20, 20, False)}")
        actual = df.head().asDict()[column]
        assert actual == expected



import pytest

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col, lit

from koheesio.logger import LoggingFactory
from koheesio.spark.delta import DeltaTableStep
from koheesio.spark.etl_task import EtlTask
from koheesio.spark.readers.delta import DeltaTableReader, DeltaTableStreamReader
from koheesio.spark.readers.dummy import DummyReader
from koheesio.spark.transformations.sql_transform import SqlTransform
from koheesio.spark.transformations.transform import Transform
from koheesio.spark.utils import SPARK_MINOR_VERSION
from koheesio.spark.writers.delta import DeltaTableStreamWriter, DeltaTableWriter
from koheesio.spark.writers.dummy import DummyWriter

pytestmark = pytest.mark.spark


def dummy_function(df: DataFrame):
    return df.withColumn("hello", lit("world"))


def dummy_function2(df: DataFrame, name: str):
    return df.withColumn("name", lit(name))


def test_dummy_task(spark):
    my_etl_logger = LoggingFactory.get_logger(name="dummy_etl_task")
    my_etl_logger.setLevel("DEBUG")
    dummy_task = EtlTask(
        source=DummyReader(range=15),
        target=DummyWriter(vertical=False),
        transformations=[Transform(dummy_function), Transform(dummy_function2, name="pari")],
    )
    my_etl_logger.debug("Testing DEBUG log level")
    results = dummy_task.execute().target_df
    actual = results.head().asDict()
    expected = {"hello": "world", "id": 0, "name": "pari"}
    my_etl_logger.info("Testing INFO log level")

    assert actual == expected
    assert results.count() == 15
    assert dummy_task.etl_date is not None


def test_delta_task(spark):
    delta_table = DeltaTableStep(table="delta_table")
    DummyReader(range=5).read().write.format("delta").mode("append").saveAsTable("delta_table")

    delta_task = EtlTask(
        source=DeltaTableReader(table=delta_table),
        target=DeltaTableWriter(table="delta_table_out"),
        transformations=[
            SqlTransform(
                sql="SELECT ${field} FROM ${table_name} WHERE id = 0",
                table_name="delta_table_out",
                params={"field": "id"},
            ),
            Transform(dummy_function2, name="pari"),
        ],
    )
    delta_task.run()

    results = spark.table("delta_table_out").orderBy(col("id")).limit(1)

    actual = results.head().asDict()
    expected = {"id": 0, "name": "pari"}
    assert actual == expected


def test_delta_stream_task(spark, checkpoint_folder):
    from koheesio.spark.utils.connect import is_remote_session

    delta_table = DeltaTableStep(table="delta_stream_table")
    DummyReader(range=5).read().write.format("delta").mode("append").saveAsTable("delta_stream_table")
    writer = DeltaTableStreamWriter(table="delta_stream_table_out", checkpoint_location=checkpoint_folder)
    transformations = [
        SqlTransform(
            sql="SELECT ${field} FROM ${table_name} WHERE id = 0",
            table_name="temp_view",
            field="id",
        ),
        Transform(dummy_function2, name="pari"),
    ]

    delta_task = EtlTask(
        source=DeltaTableStreamReader(table=delta_table),
        target=writer,
        transformations=transformations,
    )

    if 3.4 < SPARK_MINOR_VERSION < 4.0 and is_remote_session():
        with pytest.raises(RuntimeError) as excinfo:
            delta_task.run()

        assert "https://issues.apache.org/jira/browse/SPARK-45957" in str(excinfo.value.args[0])
    else:
        delta_task.run()
        writer.streaming_query.awaitTermination(timeout=20)  # type: ignore

        out_df = spark.table("delta_stream_table_out")
        actual = out_df.head().asDict()
        expected = {"id": 0, "name": "pari"}
        assert actual == expected


def test_transformations_alias(spark: SparkSession) -> None:
    my_etl_logger = LoggingFactory.get_logger(name="transformations_alias")
    my_etl_logger.setLevel("DEBUG")
    alias_task = EtlTask(
        source=DummyReader(range=15),
        target=DummyWriter(vertical=False),
        transforms=[Transform(dummy_function), Transform(dummy_function2, name="mias")],
    )
    my_etl_logger.debug("Testing DEBUG log level")
    results = alias_task.execute().target_df
    actual = results.head().asDict()
    expected = {"hello": "world", "id": 0, "name": "mias"}
    my_etl_logger.info("Testing INFO log level")

    assert actual == expected
    assert results.count() == 15
    assert alias_task.etl_date is not None



from datetime import datetime
from pathlib import Path, PurePath

import pytest

from pyspark.sql.functions import lit

from koheesio.integrations.spark.tableau.hyper import (
    NOT_NULLABLE,
    NULLABLE,
    HyperFileDataFrameWriter,
    HyperFileListWriter,
    HyperFileParquetWriter,
    HyperFileReader,
    SqlType,
    TableDefinition,
    TableName,
)

pytestmark = pytest.mark.spark


class TestHyper:
    @pytest.fixture()
    def parquet_file(self, data_path):
        path = f"{data_path}/readers/parquet_file"
        return Path(path).glob("**/*.parquet")

    @pytest.fixture()
    def hyper_file(self, data_path):
        return f"{data_path}/readers/hyper_file/dummy.hyper"

    def test_hyper_file_reader(self, hyper_file):
        df = (
            HyperFileReader(
                path=hyper_file,
            )
            .execute()
            .df
        )

        assert df.count() == 3
        assert df.dtypes == [("string", "string"), ("int", "int"), ("timestamp", "timestamp")]

    def test_hyper_file_list_writer(self, spark):
        hw = HyperFileListWriter(
            name="test",
            table_definition=TableDefinition(
                table_name=TableName("Extract", "Extract"),
                columns=[
                    TableDefinition.Column(name="string", type=SqlType.text(), nullability=NOT_NULLABLE),
                    TableDefinition.Column(name="int", type=SqlType.int(), nullability=NULLABLE),
                    TableDefinition.Column(name="timestamp", type=SqlType.timestamp(), nullability=NULLABLE),
                ],
            ),
            data=[
                ["text_1", 1, datetime(2024, 1, 1, 0, 0, 0, 0)],
                ["text_2", 2, datetime(2024, 1, 2, 0, 0, 0, 0)],
                ["text_3", None, None],
            ],
        ).execute()

        df = (
            HyperFileReader(
                path=PurePath(hw.hyper_path),
            )
            .execute()
            .df
        )

        assert df.count() == 3
        assert df.dtypes == [("string", "string"), ("int", "int"), ("timestamp", "timestamp")]

    def test_hyper_file_parquet_writer(self, data_path, parquet_file):
        hw = HyperFileParquetWriter(
            name="test",
            table_definition=TableDefinition(
                table_name=TableName("Extract", "Extract"),
                columns=[
                    TableDefinition.Column(name="string", type=SqlType.text(), nullability=NOT_NULLABLE),
                    TableDefinition.Column(name="int", type=SqlType.int(), nullability=NULLABLE),
                    TableDefinition.Column(name="timestamp", type=SqlType.timestamp(), nullability=NULLABLE),
                ],
            ),
            files=parquet_file,
        ).execute()

        df = HyperFileReader(path=PurePath(hw.hyper_path)).execute().df

        assert df.count() == 6
        assert df.dtypes == [("string", "string"), ("int", "int"), ("timestamp", "timestamp")]

    def test_hyper_file_dataframe_writer(self, data_path, df_with_all_types):
        df = df_with_all_types.drop("void", "byte", "binary", "array", "map", "float")
        df2 = df.select([lit(None).alias(col) for col in df.columns])
        df = df.union(df2)
        df = df.withColumn("decimal", df["decimal"].cast("decimal(28,3)"))

        hw = HyperFileDataFrameWriter(
            name="test",
            df=df.drop("void", "byte", "binary", "array", "map", "float"),
        ).execute()

        df = HyperFileReader(path=PurePath(hw.hyper_path)).execute().df

        assert df.count() == 2
        assert df.dtypes == [
            ("short", "smallint"),
            ("integer", "int"),
            ("long", "bigint"),
            ("double", "float"),
            ("decimal", "decimal(18,5)"),
            ("string", "string"),
            ("boolean", "boolean"),
            ("timestamp", "timestamp"),
            ("date", "date"),
        ]

    @pytest.fixture()
    def hyper_file_writer(self):
        return HyperFileListWriter(
            name="test",
            table_definition=TableDefinition(
                table_name=TableName("Extract", "Extract"),
                columns=[
                    TableDefinition.Column(name="string", type=SqlType.text(), nullability=NOT_NULLABLE),
                ],
            ),
            data=[["text_1"]],
        )

    def test_hyper_file_process_custom_log_dir(self, hyper_file_writer):
        import os
        import tempfile

        with tempfile.TemporaryDirectory() as temp_dir:
            hyper_file_writer.hyper_process_parameters = {"log_dir": temp_dir}
            hyper_file_writer.execute()

            assert os.path.exists(f"{temp_dir}/hyperd.log")



"""
Splits the contents of a column on basis of a split_pattern

Classes
-------
SplitAll
    Splits the contents of a column on basis of a split_pattern.

SplitAtFirstMatch
    Like SplitAll, but only splits the string once. You can specify whether you want the first or second part.
"""

from typing import Optional

from pyspark.sql import Column
from pyspark.sql.functions import coalesce, lit, split

from koheesio.models import Field
from koheesio.spark.transformations import ColumnsTransformationWithTarget


class SplitAll(ColumnsTransformationWithTarget):
    """
    This function splits the contents of a column on basis of a split_pattern.

    It splits at al the locations the pattern is found. The new column will be of ArrayType.

    Wraps the pyspark.sql.functions.split function.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to split. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    split_pattern : str
        This is the pattern that will be used to split the column contents.

    Example
    -------

    ### Splitting with a space as a pattern:

    __input_df:__

    |            product|amount|country|
    |-------------------|------|-------|
    |Banana lemon orange|  1000|    USA|
    |Carrots Blueberries|  1500|    USA|
    |              Beans|  1600|    USA|

    ```python
    output_df = SplitColumn(
        column="product", target_column="split", split_pattern=" "
    ).transform(input_df)
    ```

    __output_df:__

    |product            |amount|country|split                       |
    |-------------------|------|-------|----------------------------|
    |Banana lemon orange|  1000|    USA|["Banana", "lemon" "orange"]|
    |Carrots Blueberries|  1500|    USA|["Carrots", "Blueberries"]  |
    |              Beans|  1600|    USA|["Beans"]                   |
    """

    split_pattern: str = Field(default=..., description="The pattern to split the column contents.")

    def func(self, column: Column) -> Column:
        return split(column, pattern=self.split_pattern)


class SplitAtFirstMatch(SplitAll):
    """
    Like SplitAll, but only splits the string once. You can specify whether you want the first or second part.

    Note
    ----
    * SplitAtFirstMatch splits the string only once. To specify whether you want the first part or the second you
        can toggle the parameter retrieve_first_part.
    * The new column will be of StringType.
    * If you want to split a column more than once, you should call this function multiple times.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to split. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    split_pattern : str
        This is the pattern that will be used to split the column contents.
    retrieve_first_part : Optional[bool], optional, default=True
        Takes the first part of the split when true, the second part when False. Other parts are ignored.
        Defaults to True.

    Example
    -------

    ### Splitting with a space as a pattern:

    __input_df:__

    |product            |amount|country|
    |-------------------|------|-------|
    |Banana lemon orange|  1000|    USA|
    |Carrots Blueberries|  1500|    USA|
    |              Beans|  1600|    USA|

    ```python
    output_df = SplitColumn(
        column="product", target_column="split_first", split_pattern="an"
    ).transform(input_df)
    ```

    __output_df:__

    |product            |amount|country|split_first        |
    |-------------------|------|-------|-------------------|
    |Banana lemon orange|  1000|    USA|B                  |
    |Carrots Blueberries|  1500|    USA|Carrots Blueberries|
    |              Beans|  1600|    USA|Be                 |
    """

    retrieve_first_part: Optional[bool] = Field(
        default=True,
        description="Takes the first part of the split when true, the second part when False. Other parts are ignored.",
    )

    def func(self, column: Column) -> Column:
        split_func = split(column, pattern=self.split_pattern)

        # first part
        if self.retrieve_first_part:
            return split_func.getItem(0)

        # or, second part
        return coalesce(split_func.getItem(1), lit(""))



"""
String replacements without using regular expressions.
"""

from typing import Optional

from pyspark.sql import Column
from pyspark.sql.functions import lit, when

from koheesio.models import Field, field_validator
from koheesio.spark.transformations import ColumnsTransformationWithTarget


class Replace(ColumnsTransformationWithTarget):
    """
    Replace all instances of a string in a column with another string.

    This transformation uses PySpark when().otherwise() functions.

    Notes
    -----
    - If original_value is not set, the transformation will replace all null values with new_value
    - If original_value is set, the transformation will replace all values matching original_value with new_value
    - Numeric values are supported, but will be cast to string in the process
    - Replace is meant for simple string replacements. If more advanced replacements are needed, use the `RegexpReplace`
        transformation instead.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to replace values in. Alias: column
    target_column : Optional[str], optional, default=None
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    original_value : Optional[str], optional, default=None
        The original value that needs to be replaced. Alias: from
    new_value : str
        The new value to replace this with. Alias: to

    Examples
    --------
    __input_df:__

    | column |
    |--------|
    | hello  |
    | world  |
    | None   |

    ### Replace all null values with a new value
    ```python
    output_df = Replace(
        column="column",
        target_column="replaced_column",
        original_value=None,  # This is the default value, so it can be omitted
        new_value="programmer",
    ).transform(input_df)
    ```

    __output_df:__

    | column | replaced_column |
    |--------|-----------------|
    | hello  | hello           |
    | world  | world           |
    | None   | programmer      |

    ### Replace all instances of a string in a column with another string
    ```python
    output_df = Replace(
        column="column",
        target_column="replaced_column",
        original_value="world",
        new_value="programmer",
    ).transform(input_df)
    ```

    __output_df:__

    | column | replaced_column |
    |--------|-----------------|
    | hello  | hello           |
    | world  | programmer      |
    | None   | None            |

    """

    original_value: Optional[str] = Field(
        default=None, alias="from", description="The original value that needs to be replaced"
    )
    new_value: str = Field(default=..., alias="to", description="The new value to replace this with")

    @field_validator("original_value", "new_value", mode="before")
    def cast_values_to_str(cls, value: Optional[str]) -> Optional[str]:
        """Cast values to string if they are not None"""
        if value:
            return str(value)

    def func(self, column: Column) -> Column:
        when_statement = (
            when(column.isNull(), lit(self.new_value))
            if not self.original_value
            else when(
                column == self.original_value,
                lit(self.new_value),
            )
        )
        return when_statement.otherwise(column)



import math

import pytest

from pyspark.sql.types import (
    ArrayType,
    FloatType,
    IntegerType,
    StringType,
    StructField,
    StructType,
)

from koheesio.spark.transformations.arrays import (
    ArrayDistinct,
    ArrayMax,
    ArrayMean,
    ArrayMedian,
    ArrayMin,
    ArrayRemove,
    ArrayReverse,
    ArraySortAsc,
    ArraySortDesc,
    ArraySum,
    Explode,
    ExplodeDistinct,
)

pytestmark = pytest.mark.spark

input_data = [
    (1, [1, 2, 2, 3, 3, 3], ["a", "b", "b", "c", "c", " ", None], [1.0, 2.0, 2.0, 3.0, 3.0, 3.0, float("nan")]),
    (2, [4, 4, 5, 5, 6, 6, None], ["d", "e", "e", "f", "f", "f"], [4.0, 5.0, 5.0, 6.0, 6.0, 6.0]),
    (3, [7, 7, 8, 8, 9, 9], ["g", "h", "h", "i", "i", "i"], [7.0, 8.0, 8.0, 9.0, 9.0, 9.0]),
    (4, [], [], []),
    (5, None, None, None),
    (6, [-1, -3, -5, 0, +1], [], []),
]
input_schema = StructType(
    [
        StructField("id", IntegerType(), True),
        StructField("array_int", ArrayType(IntegerType()), True),
        StructField("array_str", ArrayType(StringType()), True),
        StructField("array_float", ArrayType(FloatType()), True),
    ]
)


@pytest.mark.parametrize(
    "kls,column,params,expected_data",
    [
        (
            # Explode - test with strings and preserve_nulls False
            Explode,
            "array_str",
            {"preserve_nulls": False},
            ["a", "b", "b", "c", "c", " ", None, "d", "e", "e", "f", "f", "f", "g", "h", "h", "i", "i", "i"],
        ),
        (
            # Explode - test with strings and preserve_nulls True
            Explode,
            "array_str",
            {"preserve_nulls": True},
            [
                "a",
                "b",
                "b",
                "c",
                "c",
                " ",
                None,
                "d",
                "e",
                "e",
                "f",
                "f",
                "f",
                "g",
                "h",
                "h",
                "i",
                "i",
                "i",
                None,
                None,
                None,
            ],
        ),
        (
            # Explode - test with numbers
            Explode,
            "array_int",
            None,  # no params
            [1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 6, 6, None, 7, 7, 8, 8, 9, 9, None, None, -1, -3, -5, 0, 1],
        ),
        (
            # Explode - test with numbers and preserve_nulls False
            Explode,
            "array_int",
            {"preserve_nulls": False},
            [1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 6, 6, None, 7, 7, 8, 8, 9, 9, -1, -3, -5, 0, 1],
        ),
        (
            # ArrayDistinct - test with strings
            ArrayDistinct,
            "array_str",
            None,  # no params
            [["a", "b", "c"], ["d", "e", "f"], ["g", "h", "i"], [], None, []],
        ),
        (
            # ArrayDistinct - test with numbers [int]
            ArrayDistinct,
            "array_int",
            None,  # no params
            [[1, 2, 3], [4, 5, 6], [7, 8, 9], [], None, [-1, -3, -5, 1]],
        ),
        (
            # ArrayDistinct - test with numbers [float]
            ArrayDistinct,
            "array_float",
            None,  # no params
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0], [], None, []],
        ),
        (
            # ExplodeDistinct - test with strings
            ExplodeDistinct,
            "array_str",
            None,  # no params
            ["a", "b", "c", "d", "e", "f", "g", "h", "i", None, None, None],
        ),
        (
            # ArrayReverse - test with strings
            ArrayReverse,
            "array_str",
            None,  # no params
            [
                [None, " ", "c", "c", "b", "b", "a"],
                ["f", "f", "f", "e", "e", "d"],
                ["i", "i", "i", "h", "h", "g"],
                [],
                None,
                [],
            ],
        ),
        (
            # ArraySortAsc - test with strings
            ArraySortAsc,
            "array_str",
            None,  # no params
            [
                [" ", "a", "b", "b", "c", "c", None],
                ["d", "e", "e", "f", "f", "f"],
                ["g", "h", "h", "i", "i", "i"],
                [],
                None,
                [],
            ],
        ),
        (
            # ArraySortDesc - test with strings
            ArraySortDesc,
            "array_str",
            None,  # no params
            [
                [None, "c", "c", "b", "b", "a", " "],
                ["f", "f", "f", "e", "e", "d"],
                ["i", "i", "i", "h", "h", "g"],
                [],
                None,
                [],
            ],
        ),
        (
            # ArrayRemove - test with strings
            ArrayRemove,
            "array_str",
            {"value": "b", "make_distinct": False},
            [["a", "c", "c", " "], ["d", "e", "e", "f", "f", "f"], ["g", "h", "h", "i", "i", "i"], [], None, []],
        ),
        (
            # ArrayRemove - test with multiple values
            ArrayRemove,
            "array_str",
            {"value": ["b", "f", "i"], "make_distinct": True},
            [["a", "c", " "], ["d", "e"], ["g", "h"], [], None, []],
        ),
        (
            ArrayRemove,
            "array_float",
            {"value": 2.0, "make_distinct": False},
            [
                [1.0, 3.0, 3.0, 3.0],
                [4.0, 5.0, 5.0, 6.0, 6.0, 6.0],
                [7.0, 8.0, 8.0, 9.0, 9.0, 9.0],
                [],
                None,
                [],
            ],
        ),
        (
            ArrayRemove,
            "array_float",
            {"value": 2.0, "make_distinct": False, "keep_nan": True},
            [
                [1.0, 3.0, 3.0, 3.0, float("nan")],
                [4.0, 5.0, 5.0, 6.0, 6.0, 6.0],
                [7.0, 8.0, 8.0, 9.0, 9.0, 9.0],
                [],
                None,
                [],
            ],
        ),
        (
            ArrayRemove,
            "array_int",
            {"value": 6, "make_distinct": False},
            [[1, 2, 2, 3, 3, 3], [4, 4, 5, 5], [7, 7, 8, 8, 9, 9], [], None, [-1, -3, -5, 0, +1]],
        ),
        (
            ArrayRemove,
            "array_int",
            {"value": 6, "make_distinct": False, "keep_null": True},
            [[1, 2, 2, 3, 3, 3], [4, 4, 5, 5, None], [7, 7, 8, 8, 9, 9], [], None, [-1, -3, -5, 0, +1]],
        ),
        (
            # ArrayMin - test with numbers [int]
            ArrayMin,
            "array_int",
            None,  # no params
            [1, 4, 7, None, None, -5],
        ),
        (
            # ArrayMin - test with numbers [float]
            ArrayMin,
            "array_float",
            None,  # no params
            [1.0, 4.0, 7.0, None, None, None],
        ),
        (
            # ArrayMin - test with strings
            ArrayMin,
            "array_str",
            None,  # no params
            [" ", "d", "g", None, None, None],
        ),
        (
            # ArrayMax - test with numbers [int]
            ArrayMax,
            "array_int",
            None,  # no params
            [3, 6, 9, None, None, 1],
        ),
        (
            # ArrayMax - test with numbers [float]
            ArrayMax,
            "array_float",
            None,  # no params
            [3.0, 6.0, 9.0, None, None, None],
        ),
        (
            # ArrayMax - test with numbers [float]
            ArrayMax,
            "array_float",
            {"keep_nan": True},
            [float("nan"), 6.0, 9.0, None, None, None],
        ),
        (
            # ArrayMax - test with strings
            ArrayMax,
            "array_str",
            None,  # no params
            ["c", "f", "i", None, None, None],
        ),
        (
            # ArraySum - test with numbers [int]
            ArraySum,
            "array_int",
            None,  # no params
            [14, 30, 48, 0, None, -8],
        ),
        (
            # ArraySum - test with numbers [float]
            ArraySum,
            "array_float",
            None,  # no params
            [14.0, 32.0, 50.0, 0.0, None, 0.0],
        ),
        (
            # ArrayMean - test with numbers [int]
            ArrayMean,
            "array_int",
            None,  # no params
            [2.3333333333333335, 5.0, 8.0, 0.0, None, -1.6],
        ),
        (
            # ArrayMean - test with numbers [int]
            ArrayMean,
            "array_int",
            {"keep_null": True},
            [2.3333333333333335, None, 8.0, 0.0, None, -1.6],
        ),
        (
            # ArrayMean - test with numbers [float]
            ArrayMean,
            "array_float",
            None,  # no params
            [2.3333333333333335, 5.333333333333333, 8.333333333333334, 0.0, None, 0.0],
        ),
        (
            # ArrayMean - test with numbers [float]
            ArrayMean,
            "array_float",
            {"keep_nan": True},
            [float("nan"), 5.333333333333333, 8.333333333333334, 0.0, None, 0.0],
        ),
        (
            # ArrayMedian - test with numbers [int]
            ArrayMedian,
            "array_int",
            None,  # no params
            [2.5, 5.0, 8.0, 0.0, None, -1.0],
        ),
        (
            # ArrayMedian - test with numbers [float]
            ArrayMedian,
            "array_float",
            None,  # no params
            [2.5, 5.5, 8.5, 0.0, None, 0.0],
        ),
    ],
)
def test_array(kls, column, expected_data, params, spark):
    # set up test data
    test_data = spark.createDataFrame(
        data=input_data,
        schema=input_schema,
    )

    params = params or {}

    # noinspection PyCallingNonCallable
    df = kls(df=test_data, column=column, **params).transform()
    actual_data = [row.asDict()[column] for row in df.select(column).collect()]

    def check_result(_actual_data: list, _expected_data: list):
        _data = _expected_data or _actual_data

        if _data:
            for i in range(len(_data)):
                if isinstance(_actual_data[i], float):
                    if math.isnan(_expected_data[i]):
                        assert math.isnan(_actual_data[i])
                    else:
                        assert _actual_data[i] == _expected_data[i]
                else:
                    assert _actual_data[i] == _expected_data[i]
        else:
            assert _actual_data == _expected_data

    if isinstance(expected_data[0], list):
        for actual, expected in zip(actual_data, expected_data):
            check_result(actual, expected)
    else:
        check_result(actual_data, expected_data)


@pytest.mark.parametrize(
    "column,params,expected_type",
    [
        ("array_int", None, "int"),
        ("array_int", {"distinct": True}, "int"),
        ("array_float", None, "float"),
        ("array_float", {"distinct": True}, "float"),
    ],
)
def test_explode_type_inherited(spark, column, params, expected_type):
    """Type of the target column should be inherited by the array type"""
    test_data = spark.createDataFrame(
        data=input_data,
        schema=input_schema,
    )

    params = params or {}

    result = Explode(df=test_data, column=column, target_column="exploded", **params).transform()
    result.show()
    result_type = result.select("exploded").dtypes[0][1]

    assert result_type == expected_type



"""Module for secret integrations.

Contains abstract class for various secret integrations also known as SecretContext.
"""

from typing import Optional
from abc import ABC, abstractmethod

from koheesio import Step, StepOutput
from koheesio.context import Context
from koheesio.models import Field, SecretStr


class Secret(Step, ABC):
    """
    Abstract class for various secret integrations.
    All secrets are wrapped into Context class for easy access. Either existing context can be provided,
    or new context will be created and returned at runtime.

    Secrets are wrapped into the pydantic.SecretStr.
    """

    context: Optional[Context] = Field(
        Context({}),
        description="Existing `Context` instance can be used for secrets, otherwise new empty context will be created.",
    )
    root: Optional[str] = Field(default="secrets", description="All secrets will be grouped under this root.")
    parent: Optional[str] = Field(
        default=...,
        description="Group secrets from one secure path under this friendly name",
        pattern=r"^[a-zA-Z0-9_]+$",
    )  # Enforce regex to leverage the property style notation while working with Context: root.parent.secret

    class Output(StepOutput):
        """Output class for Secret."""

        context: Context = Field(default=..., description="Koheesio context")

    @classmethod
    def encode_secret_values(cls, data: dict) -> dict:
        """Encode secret values in the dictionary.

        Ensures that all values in the dictionary are wrapped in SecretStr.
        """
        encoded_dict = {}
        for key, value in data.items():
            if isinstance(value, dict):
                encoded_dict[key] = cls.encode_secret_values(value)
            else:
                encoded_dict[key] = SecretStr(value)  # type: ignore[assignment]
        return encoded_dict

    @abstractmethod
    def _get_secrets(self) -> dict:
        """
        Implement this method to return dictionary of protected secrets.
        """
        ...

    def execute(self) -> None:
        """
        Main method to handle secrets protection and context creation with "root-parent-secrets" structure.
        """
        context = Context(self.encode_secret_values(data={self.root: {self.parent: self._get_secrets()}}))
        self.output.context = self.context.merge(context=context)

    # noinspection PyMethodOverriding
    def get(self) -> Context:
        """
        Convenience method to return context with secrets.
        """
        self.execute()
        return self.output.context



import warnings

from aiohttp import ClientResponseError, ClientSession, TCPConnector
from aiohttp_retry import ExponentialRetry
import pytest
from yarl import URL

from pydantic import ValidationError

from koheesio.asyncio.http import AsyncHttpStep
from koheesio.steps.http import HttpMethod

# noinspection HttpUrlsUsage
ASYNC_BASE_URL = "http://httpbin.org"
ASYNC_GET_ENDPOINT = URL(f"{ASYNC_BASE_URL}/get")
ASYNC_STATUS_503_ENDPOINT = URL(f"{ASYNC_BASE_URL}/status/503")
ASYNC_STATUS_404_ENDPOINT = URL(f"{ASYNC_BASE_URL}/status/404")


def test_async_http_get_step_positive():
    """
    Testing the GET function with a positive scenario.
    """
    step = AsyncHttpStep(
        method=HttpMethod.GET,
        url=[
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
            ASYNC_GET_ENDPOINT,
        ],
    )
    step.execute()
    responses_urls = step.output.responses_urls

    assert len(responses_urls) == 10
    response, url = responses_urls[0]
    assert url == ASYNC_GET_ENDPOINT
    assert response["url"] == str(ASYNC_GET_ENDPOINT)
    assert response["headers"]["Host"] == "httpbin.org"


@pytest.mark.parametrize(
    "http_method, status_endpoint, expected_status",
    [
        (HttpMethod.GET, ASYNC_STATUS_503_ENDPOINT, 503),
        (HttpMethod.GET, ASYNC_STATUS_404_ENDPOINT, 404),
        (HttpMethod.POST, ASYNC_STATUS_503_ENDPOINT, 503),
        (HttpMethod.POST, ASYNC_STATUS_404_ENDPOINT, 404),
        (HttpMethod.PUT, ASYNC_STATUS_503_ENDPOINT, 503),
        (HttpMethod.PUT, ASYNC_STATUS_404_ENDPOINT, 404),
        (HttpMethod.DELETE, ASYNC_STATUS_503_ENDPOINT, 503),
        (HttpMethod.DELETE, ASYNC_STATUS_404_ENDPOINT, 404),
    ],
)
def test_async_http_step_negative(http_method, status_endpoint, expected_status):
    """
    Testing the function with a negative scenario (503 and 404 status codes).
    """
    step = AsyncHttpStep(method=http_method, url=[status_endpoint])
    with pytest.raises(ClientResponseError) as excinfo:
        step.execute()

    assert excinfo.value.status == expected_status


@pytest.mark.asyncio
async def test_async_http_step():
    """
    Testing the AsyncHttpStep class.
    """
    # Initialize the AsyncHttpStep
    session = ClientSession()
    urls = [URL(ASYNC_GET_ENDPOINT), URL(ASYNC_GET_ENDPOINT)]
    retry_options = ExponentialRetry()
    connector = TCPConnector(limit=10)
    headers = {"Content-Type": "application/json"}
    step = AsyncHttpStep(
        client_session=session, url=urls, retry_options=retry_options, connector=connector, headers=headers
    )

    # Execute the step
    responses_urls = await step.get()

    # Assert the responses_urls
    assert isinstance(responses_urls, list)
    assert len(responses_urls) == 2


@pytest.mark.asyncio
async def test_async_http_step_with_timeout():
    """
    Testing the AsyncHttpStep class with timeout.
    """
    # Initialize the AsyncHttpStep
    session = ClientSession()
    urls = [URL(ASYNC_GET_ENDPOINT), URL(ASYNC_GET_ENDPOINT)]
    retry_options = ExponentialRetry()
    connector = TCPConnector(limit=10)
    headers = {"Content-Type": "application/json"}

    # Assert the warning
    with pytest.raises(ValidationError):
        AsyncHttpStep(
            client_session=session,
            url=urls,
            retry_options=retry_options,
            connector=connector,
            headers=headers,
            timeout=10,
        )


def test_async_http_step_with_invalid_http_method():
    """
    Testing the function with an invalid HTTP method.
    """
    invalid_method = "INVALID_METHOD"

    with pytest.raises(ValueError) as exc_info:
        step = AsyncHttpStep(method=invalid_method, url=[ASYNC_GET_ENDPOINT])
        step.execute()

    assert str(exc_info.value) == f"Method {invalid_method} not implemented in AsyncHttpStep."


@pytest.mark.asyncio
async def test_async_http_step_set_outputs_warning():
    """
    Testing the AsyncHttpStep class's set_outputs method for warning.
    """
    # Initialize the AsyncHttpStep
    session = ClientSession()
    urls = [URL(ASYNC_GET_ENDPOINT), URL(ASYNC_GET_ENDPOINT)]
    retry_options = ExponentialRetry()
    connector = TCPConnector(limit=10)
    headers = {"Content-Type": "application/json"}
    step = AsyncHttpStep(
        client_session=session,
        url=urls,
        retry_options=retry_options,
        connector=connector,
        headers=headers,
    )

    # Assert the warning
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        step.set_outputs(None)
        # Verify some things
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert "set outputs is not implemented in AsyncHttpStep." == str(w[-1].message)


@pytest.mark.asyncio
async def test_async_http_step_get_options_warning():
    """
    Testing the AsyncHttpStep class's get_options method for warning.
    """
    # Initialize the AsyncHttpStep
    session = ClientSession()
    urls = [URL(ASYNC_GET_ENDPOINT), URL(ASYNC_GET_ENDPOINT)]
    retry_options = ExponentialRetry()
    connector = TCPConnector(limit=10)
    headers = {"Content-Type": "application/json"}
    step = AsyncHttpStep(
        client_session=session,
        url=urls,
        retry_options=retry_options,
        connector=connector,
        headers=headers,
    )

    # Assert the warning
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        step.get_options()
        # Verify some things
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert "get_options is not implemented in AsyncHttpStep." == str(w[-1].message)



from datetime import datetime

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.date_time import (
    ChangeTimeZone,
    DateFormat,
    ToTimestamp,
)

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name="test_date_time")


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: base test
            # input values,
            dict(target_column="a_date", column="a_date", format="yyyyMMdd HH:mm"),
            # expected data
            [
                dict(id=1, a_date="19700420 12:33", a_timestamp=datetime(2000, 7, 1, 1, 1)),
                dict(id=2, a_date="19800521 13:34", a_timestamp=datetime(2010, 8, 2, 2, 2)),
                dict(id=3, a_date="19900622 14:35", a_timestamp=datetime(2020, 9, 3, 3, 3)),
            ],
        ),
        (
            # description: giving a list of the same columns names should not cause an error
            # input values,
            dict(target_column="a_date", columns=["a_date", "a_date", "a_date"], format="yyyyMMdd HH:mm"),
            # expected data
            [
                dict(id=1, a_date="19700420 12:33", a_timestamp=datetime(2000, 7, 1, 1, 1)),
                dict(id=2, a_date="19800521 13:34", a_timestamp=datetime(2010, 8, 2, 2, 2)),
                dict(id=3, a_date="19900622 14:35", a_timestamp=datetime(2020, 9, 3, 3, 3)),
            ],
        ),
        (
            # description: multiple columns at once, no target_columns -> should output to original column name
            # input values,
            dict(columns=["a_date", "a_timestamp"], format="yyyyMMdd HH:mm"),
            # expected data
            [
                dict(id=1, a_date="19700420 12:33", a_timestamp="20000701 01:01"),
                dict(id=2, a_date="19800521 13:34", a_timestamp="20100802 02:02"),
                dict(id=3, a_date="19900622 14:35", a_timestamp="20200903 03:03"),
            ],
        ),
        (
            # description: multiple columns at once, with target_columns -> should add suffix to the column name
            # input values,
            dict(target_column="add", columns=["a_date", "a_timestamp"], format="yyyyMMdd HH:mm"),
            # expected data
            [
                dict(
                    id=1,
                    a_date=datetime(1970, 4, 20, 12, 33, 9),
                    a_date_add="19700420 12:33",
                    a_timestamp=datetime(2000, 7, 1, 1, 1),
                    a_timestamp_add="20000701 01:01",
                ),
                dict(
                    id=2,
                    a_date=datetime(1980, 5, 21, 13, 34, 8),
                    a_date_add="19800521 13:34",
                    a_timestamp=datetime(2010, 8, 2, 2, 2),
                    a_timestamp_add="20100802 02:02",
                ),
                dict(
                    id=3,
                    a_date=datetime(1990, 6, 22, 14, 35, 7),
                    a_date_add="19900622 14:35",
                    a_timestamp=datetime(2020, 9, 3, 3, 3),
                    a_timestamp_add="20200903 03:03",
                ),
            ],
        ),
        (
            # description: multiple columns at once, overlapping target_columns
            #               -> should add 'a_date' as suffix to the 'a_timestamp' column, while overwriting 'a_date'
            # input values,
            dict(target_column="a_date", columns=["a_date", "a_timestamp"], format="yyyyMMdd HH:mm"),
            # expected data
            [
                dict(
                    id=1,
                    a_date="19700420 12:33",
                    a_timestamp=datetime(2000, 7, 1, 1, 1),
                    a_timestamp_a_date="20000701 01:01",
                ),
                dict(
                    id=2,
                    a_date="19800521 13:34",
                    a_timestamp=datetime(2010, 8, 2, 2, 2),
                    a_timestamp_a_date="20100802 02:02",
                ),
                dict(
                    id=3,
                    a_date="19900622 14:35",
                    a_timestamp=datetime(2020, 9, 3, 3, 3),
                    a_timestamp_a_date="20200903 03:03",
                ),
            ],
        ),
    ],
)
def test_date_format(input_values, expected, sample_df_with_timestamp):
    df = DateFormat(**input_values).transform(sample_df_with_timestamp)
    actual = [k.asDict() for k in df.collect()]
    assert actual == expected
    log.info("DatetimeToFormattedString unit test completed")


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: base test
            # input values
            dict(
                target_column="a_date",
                column="a_timestamp",
                from_timezone="CET",
                to_timezone="UTC",
                add_target_timezone=False,
            ),
            # expected data
            [
                dict(id=1, a_date=datetime(2000, 6, 30, 23, 1, 0), a_timestamp=datetime(2000, 7, 1, 1, 1)),
                dict(id=2, a_date=datetime(2010, 8, 2, 0, 2, 0), a_timestamp=datetime(2010, 8, 2, 2, 2)),
                dict(id=3, a_date=datetime(2020, 9, 3, 1, 3, 0), a_timestamp=datetime(2020, 9, 3, 3, 3)),
            ],
        ),
        (
            # description: base test
            # input values
            dict(
                target_column="a_date",
                column="a_timestamp",
                from_timezone="CET",
                to_timezone="UTC",
                add_target_timezone=True,
            ),
            # expected data
            [
                dict(
                    id=1,
                    a_date=datetime(2000, 6, 30, 23, 1, 0),
                    a_timestamp=datetime(2000, 7, 1, 1, 1),
                    a_date_timezone="UTC",
                ),
                dict(
                    id=2,
                    a_date=datetime(2010, 8, 2, 0, 2, 0),
                    a_timestamp=datetime(2010, 8, 2, 2, 2),
                    a_date_timezone="UTC",
                ),
                dict(
                    id=3,
                    a_date=datetime(2020, 9, 3, 1, 3, 0),
                    a_timestamp=datetime(2020, 9, 3, 3, 3),
                    a_date_timezone="UTC",
                ),
            ],
        ),
    ],
)
def test_change_time_zone(input_values, expected, sample_df_with_timestamp):
    df = ChangeTimeZone(**input_values).transform(df=sample_df_with_timestamp)
    actual = [k.asDict() for k in df.collect()]
    assert actual == expected
    log.info("ChangeTimeZone unit test completed")


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: base test
            # input values
            dict(
                target_column="a_timestamp",
                column="a_string_timestamp",
                format="yyyyMMddHHmm",
            ),
            # expected data
            [
                dict(id=1, a_string_timestamp="202301010420", a_timestamp=datetime(2023, 1, 1, 4, 20, 0)),
                dict(id=2, a_string_timestamp="202302020314", a_timestamp=datetime(2023, 2, 2, 3, 14, 0)),
            ],
        )
    ],
)
def test_to_timestamp(input_values, expected, sample_df_with_string_timestamp):
    df = ToTimestamp(**input_values).transform(sample_df_with_string_timestamp)
    actual = [k.asDict() for k in df.collect()]
    assert actual == expected
    log.info("ToTimestamp unit test completed")



"""
Utility functions related to date and time operations
"""

from typing import Optional
from datetime import timedelta
import re

from koheesio.models import BaseModel, model_validator

extract_dt_interval = re.compile(
    r"""
    (?P<years>\d+)\s+years?\s*|
    (?P<months>\d+)\s+months?\s*|
    (?P<weeks>\d+)\s+weeks?\s*|
    (?P<days>\d+)\s+days?\s*|
    (?P<hours>\d+)\s+hours?\s*|
    (?P<minutes>\d+)\s+(?:minutes?|mins?)\s*|
    (?P<seconds>\d+)\s+(?:seconds?|secs?)\s*|
    (?P<milliseconds>\d+)\s+(?:milliseconds?|millis?)\s*|
    (?P<microseconds>\d+)\s+(?:microseconds?|micros?)\s*
""",
    re.VERBOSE,
)


class DTInterval(BaseModel):
    """
    A class to define date and time intervals using human-readable strings or individual time components.

    Parameters
    ----------
    interval : str, optional
        A human-readable string specifying the duration of the interval, broken down
        into years, months, weeks, days, hours, minutes, seconds, milliseconds, and microseconds.
    years : int, optional, default 0
        Number of years in the interval.
    months : int, optional, default 0
        Number of months in the interval.
    weeks : int, optional, default 0
        Number of weeks in the interval.
    days : int, optional, default 0
        Number of days in the interval.
    hours : int, optional, default 0
        Number of hours in the interval.
    minutes : int, optional, default 0
        Number of minutes in the interval.
    seconds : int, optional, default 0
        Number of seconds in the interval.
    milliseconds : int, optional, default 0
        Number of milliseconds in the interval.
    microseconds : int, optional, default 0
        Number of microseconds in the interval.

    Examples
    --------
    Creating an instance with time components:

    ```python
    print(DTInterval(years=2, weeks=3, hours=12).to_timedelta)
    ```
    751 days, 12:00:00

    Creating an instance from a string:
    ```python
    print(
        DTInterval(
            interval="1 year 2 months 3 weeks 4 days 5 hours 100 minutes 200 seconds 300 milliseconds 400 microseconds"
        ).to_timedelta
    )
    ```
    451 days, 6:43:20.300400

    Methods
    -------
    to_timedelta()
        Converts the DTInterval instance to a timedelta object, aggregating all specified time components.
    """

    interval: Optional[str] = None
    years: Optional[int] = 0  # = days * 365.25 (average year length, rounded up)
    months: Optional[int] = 0  # = days * 30.44 (average month length, rounded up)
    weeks: Optional[int] = 0
    days: Optional[int] = 0
    hours: Optional[int] = 0
    minutes: Optional[int] = 0
    seconds: Optional[int] = 0
    milliseconds: Optional[int] = 0
    microseconds: Optional[int] = 0

    @model_validator(mode="before")
    def process_interval(cls, values: dict) -> dict:
        """Processes the input interval string and extracts the time components"""
        if interval_value := values.get("interval"):
            matches = extract_dt_interval.finditer(interval_value)

            # update values with the extracted values
            for match in matches:
                for key, value in match.groupdict().items():
                    if value:
                        values[key] = int(value)
        return values

    @model_validator(mode="after")
    def calculate_days(self) -> "DTInterval":
        """Years and months are not supported in timedelta, so we need to convert them to days"""
        if self.years or self.months:
            year_month_days = int(self.years * 365.25 + self.months * 30.44)  # average year length
            self.days += year_month_days
            self.years = 0
            self.months = 0
        return self

    @property
    def to_timedelta(self) -> timedelta:
        """Returns the object as a timedelta object"""
        return timedelta(
            days=self.days,
            seconds=self.seconds,
            microseconds=self.microseconds,
            milliseconds=self.milliseconds,
            minutes=self.minutes,
            hours=self.hours,
            weeks=self.weeks,
        )



"""
Class for converting DataFrame column names from camel case to snake case.
"""

from typing import Optional
import re

from koheesio.models import Field, ListOfColumns
from koheesio.spark.transformations import ColumnsTransformation
from koheesio.spark.utils import SPARK_MINOR_VERSION

camel_to_snake_re = re.compile("([a-z0-9])([A-Z])")


def convert_camel_to_snake(name: str) -> str:
    """
    Converts a string from camelCase to snake_case.

    Parameters:
    ----------
    name : str
        The string to be converted.

    Returns:
    --------
    str
        The converted string in snake_case.
    """
    return camel_to_snake_re.sub(r"\1_\2", name).lower()


class CamelToSnakeTransformation(ColumnsTransformation):
    """
    Converts column names from camel case to snake cases

    Parameters
    ----------
    columns : Optional[ListOfColumns], optional, default=None
        The column or columns to convert. If no columns are specified, all columns will be converted. A list of columns
        or a single column can be specified.
        For example: `["column1", "column2"]` or `"column1"`

    Example
    -------
    __input_df:__

    | camelCaseColumn    | snake_case_column |
    |--------------------|-------------------|
    | ...                | ...               |

    ```python
    output_df = CamelToSnakeTransformation(
        column="camelCaseColumn"
    ).transform(input_df)
    ```

    __output_df:__

    | camel_case_column | snake_case_column |
    |-------------------|-------------------|
    | ...               | ...               |

    In this example, the column `camelCaseColumn` is converted to `camel_case_column`.

    > Note: the data in the columns is not changed, only the column names.

    """

    def execute(self) -> ColumnsTransformation.Output:
        _df = self.df

        # Prepare columns input:
        columns = list(self.get_columns())

        if SPARK_MINOR_VERSION < 3.4:
            for column in columns:
                _df = _df.withColumnRenamed(column, convert_camel_to_snake(column))

        else:
            # Rename columns using toDF for Spark versions >= 3.4
            # Note: toDF requires all column names to be specified
            new_column_names = []
            for column in _df.columns:
                if column in columns:
                    new_column_names.append(convert_camel_to_snake(column))
                    continue
                new_column_names.append(column)
            _df = _df.toDF(*new_column_names)

        self.output.df = _df



import datetime
from pathlib import Path

from koheesio.spark.readers.excel import ExcelReader


def test_excel_reader(spark, data_path):
    # Define the path to the test Excel file and the expected DataFrame
    test_file = Path(data_path) / "readers" / "excel_file" / "dummy.xlsx"

    # Initialize the ExcelReader with the path to the test file and the correct sheet name
    reader = ExcelReader(path=test_file, sheet_name="sheet_to_select", header=0)

    # Execute the reader
    reader.execute()

    # Define the expected DataFrame
    expected_df = spark.createDataFrame(
        [
            ("foo", "bar", "baz", None, datetime.datetime(2024, 1, 1, 0, 0)),
            ("so long", "and thanks", "for all the fish", 42, datetime.datetime(2024, 1, 1, 0, 0)),
        ],
        ["a", "b", "c", "d", "e"],
    )

    # Assert that the output DataFrame is as expected
    assert sorted(reader.output.df.collect()) == sorted(expected_df.collect())



"""
This module contains Okta integration steps.
"""

from __future__ import annotations

from typing import Dict, Optional
from logging import Filter, LogRecord

from requests import HTTPError

from koheesio.logger import LoggingFactory, MaskedString
from koheesio.models import Field, SecretStr
from koheesio.steps.http import HttpPostStep


class Okta(HttpPostStep):
    """
    Base Okta class
    """

    client_id: str = Field(default=..., alias="okta_id", description="Okta account ID")
    client_secret: SecretStr = Field(default=..., alias="okta_secret", description="Okta account secret", repr=False)
    data: Dict[str, str] = Field(
        default={"grant_type": "client_credentials"}, description="Data to be sent along with the token request"
    )

    # headers are not used in this class
    headers: dict = {}

    def get_options(self) -> dict:
        """options to be passed to requests.request()"""
        _options = super().get_options()
        _options["auth"] = (self.client_id, MaskedString(self.client_secret.get_secret_value()))
        return _options


class LoggerOktaTokenFilter(Filter):
    """Filter which hides token value from log."""

    def __init__(self, okta_object: OktaAccessToken, name: str = "OktaToken"):
        self.__okta_object = okta_object
        super().__init__(name=name)

    def filter(self, record: LogRecord) -> bool:
        # noinspection PyUnresolvedReferences
        if token := self.__okta_object.output.token:
            token_value = token.get_secret_value()
            record.msg = record.msg.replace(token_value, "<SECRET_TOKEN>")

        return True


class OktaAccessToken(Okta):
    """
    Get Okta authorization token

    Examples
    --------
    ```python
    token = (
        OktaAccessToken(
            url="https://org.okta.com",
            client_id="client",
            client_secret=SecretStr("secret"),
            params={
                "p1": "foo",
                "p2": "bar",
            },
        )
        .execute()
        .token
    )
    ```
    """

    class Output(Okta.Output):
        """Output class for OktaAccessToken."""

        token: Optional[SecretStr] = Field(default=None, description="Okta authentication token")

    def __init__(self, **kwargs):  # type: ignore[no-untyped-def]
        _logger = LoggingFactory.get_logger(name=self.__class__.__name__, inherit_from_koheesio=True)
        logger_filter = LoggerOktaTokenFilter(okta_object=self)
        _logger.addFilter(logger_filter)
        super().__init__(**kwargs)

    def execute(self) -> None:
        """
        Execute an HTTP Post call to Okta service and retrieve the access token.
        """
        super().execute()

        # noinspection PyUnresolvedReferences
        status_code = self.output.status_code
        # noinspection PyUnresolvedReferences
        raw_payload = self.output.raw_payload

        if status_code != 200:
            raise HTTPError(
                f"Request failed with '{status_code}' code. Payload: {raw_payload}",
                response=self.output.response_raw,
                request=None,
            )

        # noinspection PyUnresolvedReferences
        json_payload = self.output.json_payload

        if token := json_payload.get("access_token"):
            self.output.token = SecretStr(token)
        else:
            raise ValueError(f"No 'access_token' found in the Okta response: {json_payload}")



import os
from unittest.mock import MagicMock, patch

import pytest

from pyspark.sql.types import StringType, StructField, StructType

from koheesio.spark.utils import (
    get_column_name,
    import_pandas_based_on_pyspark_version,
    on_databricks,
    schema_struct_to_schema_str,
    show_string,
)
from koheesio.spark.utils.common import (
    check_if_pyspark_connect_is_supported,
    get_active_session,
    get_spark_minor_version,
)


class TestGetActiveSession:
    def test_unhappy_get_active_session_spark_connect(self):
        """Test that get_active_session raises an error when no active session is found when using spark connect."""
        with (
            # ensure that we are forcing the code to think that we are using spark connect
            patch(
                "koheesio.spark.utils.common.check_if_pyspark_connect_is_supported",
                return_value=True,
            ),
            # make sure that spark session is not found
            patch("pyspark.sql.SparkSession.getActiveSession", return_value=None),
        ):
            session = MagicMock(SparkSession=MagicMock(getActiveSession=MagicMock(return_value=None)))
            with patch.dict("sys.modules", {"pyspark.sql.connect.session": session}):
                with pytest.raises(
                    RuntimeError,
                    match="No active Spark session found. Please create a Spark session before using module "
                    "connect_utils. Or perform local import of the module.",
                ):
                    get_active_session()

    def test_unhappy_get_active_session(self):
        """Test that get_active_session raises an error when no active session is found."""
        with (
            patch(
                "koheesio.spark.utils.common.check_if_pyspark_connect_is_supported",
                return_value=False,
            ),
            patch("pyspark.sql.SparkSession.getActiveSession", return_value=None),
        ):
            with pytest.raises(
                RuntimeError,
                match="No active Spark session found. Please create a Spark session before using module connect_utils. "
                "Or perform local import of the module.",
            ):
                get_active_session()

    def test_get_active_session_with_spark(self, spark):
        """Test get_active_session when an active session is found"""
        session = get_active_session()
        assert session is not None


class TestCheckIfPysparkConnectIsSupported:
    def test_if_pyspark_connect_is_not_supported(self, mocker):
        """Test that check_if_pyspark_connect_is_supported returns False or raises an ImportError when pyspark connect
        is not supported irregardless of the pyspark version used."""
        mocker.patch.dict(
            "sys.modules",
            {
                "pyspark.sql.connect": None,
                "grpc": None,
            },
        )
        mocker.patch.dict(
            os.environ,
            {
                "SPARK_CONNECT_MODE_ENABLED": "0",
            },
            clear=True,
        )
        assert check_if_pyspark_connect_is_supported() is False

    @pytest.mark.parametrize("spark_minor_version", [3.3, 3.4, 3.5])
    def test_check_if_pyspark_connect_is_supported(self, spark_minor_version: float, mocker):
        """Test that check_if_pyspark_connect_is_supported returns True when pyspark connect is supported."""
        # Arrange
        mocker.patch("koheesio.spark.utils.common.SPARK_MINOR_VERSION", spark_minor_version)
        mocker.patch.dict(
            "sys.modules",
            {
                "pyspark.sql.connect.column": MagicMock(Column=MagicMock()),
                "pyspark.sql.connect": MagicMock(),
                "grpc": MagicMock(),
            },
        )
        mocker.patch.dict(os.environ, {"SPARK_REMOTE": "UNIT_TEST"})

        # Act
        actual_check = check_if_pyspark_connect_is_supported()

        # Assert
        assert os.environ["SPARK_REMOTE"] == "UNIT_TEST"
        if spark_minor_version >= 3.4:
            assert actual_check is True
        else:
            assert actual_check is False

    @pytest.mark.parametrize("spark_minor_version", [3.3, 3.4, 3.5])
    def test_pyspark_connect_set_without_partial_deps(self, spark_minor_version: float, mocker):
        """Test that check_if_pyspark_connect_is_supported raises an ImportError if pyspark connect is accessed without
        the dependencies being available"""
        # Arrange
        mocker.patch("koheesio.spark.utils.common.SPARK_MINOR_VERSION", spark_minor_version)
        mocker.patch.dict(
            "sys.modules",
            {
                "pyspark.sql.connect": MagicMock(Column=MagicMock()),  # pyspark module exists
                "grpc": None,  # but extras are not installed
            },
        )
        mocker.patch.dict(os.environ, {"SPARK_CONNECT_MODE_ENABLED": "1"})

        # Act & Assert
        if spark_minor_version < 3.4:
            assert check_if_pyspark_connect_is_supported() is False
        else:
            with pytest.raises(ImportError):
                check_if_pyspark_connect_is_supported()


def test_get_spark_minor_version():
    """Test that get_spark_minor_version returns the correctly formatted version."""
    with patch("koheesio.spark.utils.common.spark_version", "9.9.42"):
        assert get_spark_minor_version() == 9.9


def test_schema_struct_to_schema_str():
    struct_schema = StructType([StructField("a", StringType()), StructField("b", StringType())])
    val = schema_struct_to_schema_str(struct_schema)
    assert val == "a STRING,\nb STRING"
    assert schema_struct_to_schema_str(None) == ""


@pytest.mark.parametrize(
    "env_var_value, expected_result",
    [("lts_11_spark_3_scala_2.12", True), ("unit_test", True), (None, False)],
)
def test_on_databricks(env_var_value, expected_result):
    if env_var_value is not None:
        with patch.dict(os.environ, {"DATABRICKS_RUNTIME_VERSION": env_var_value}):
            assert on_databricks() == expected_result
    else:
        with patch.dict(os.environ, clear=True):
            assert on_databricks() == expected_result


@pytest.mark.parametrize(
    "spark_version, pandas_version, expected_error",
    [
        (3.3, "1.2.3", None),  # PySpark 3.3, pandas < 2, should not raise an error
        (3.4, "2.3.4", None),  # PySpark not 3.3, pandas >= 2, should not raise an error
        (3.3, "2.3.4", ImportError),  # PySpark 3.3, pandas >= 2, should raise an error
        (
            3.4,
            "1.2.3",
            ImportError,
        ),  # PySpark not 3.3, pandas < 2, should raise an error
    ],
)
def test_import_pandas_based_on_pyspark_version(spark_version, pandas_version, expected_error):
    with (
        patch(
            "koheesio.spark.utils.common.get_spark_minor_version",
            return_value=spark_version,
        ),
        patch("pandas.__version__", new=pandas_version),
    ):
        if expected_error:
            with pytest.raises(expected_error):
                import_pandas_based_on_pyspark_version()
        else:
            import_pandas_based_on_pyspark_version()  # This should not raise an error


def test_show_string(dummy_df):
    actual = show_string(dummy_df, n=1, truncate=1, vertical=False)
    assert actual == "+---+\n| id|\n+---+\n|  0|\n+---+\n"


def test_column_name():
    from pyspark.sql.functions import col

    name = "my_column"
    column = col(name)
    assert get_column_name(column) == name



import json

from requests_mock.mocker import Mocker

from koheesio.notifications.slack import (
    NotificationSeverity,
    SlackNotification,
    SlackNotificationWithSeverity,
)

URL = "https://hooks.slack.com/services/foo/bar/baz"


def test_slack_notification():
    sn = SlackNotification(url=URL, channel="#cux", message="lorem ipsum")
    with Mocker() as rm:
        rm.post(URL, text="ok", status_code=int(200))
        sn.execute()
        assert sn.output.status_code == 200
        assert json.loads(sn.data)["channel"] == "#cux"
        assert json.loads(sn.data)["attachments"][0]["blocks"][0]["text"]["text"] == "lorem ipsum"


def test_slack_notification_wo_channel():
    sn = SlackNotification(url=URL, message="lorem ipsum")
    with Mocker() as rm:
        rm.post(URL, text="ok", status_code=int(200))
        sn.execute()
        assert sn.output.status_code == 200
        assert not json.loads(sn.data).get("channel")


def test_slack_notification_w_severity():
    sn = SlackNotificationWithSeverity(
        url=URL,
        channel="#cux",
        message="lorem ipsum",
        severity=NotificationSeverity.ERROR,
        title="Title",
        environment="local",
        application="Application",
    )
    with Mocker() as rm:
        rm.post(URL, text="ok", status_code=int(200))
        sn.execute()
        assert sn.output.status_code == 200
        assert ":red_circle:" in json.loads(sn.data)["attachments"][0]["blocks"][0]["text"]["text"]
        assert "*Message:* lorem ipsum" in json.loads(sn.data)["attachments"][0]["blocks"][0]["text"]["text"]



"""
This module contains the RowNumberDedup class, which performs a row_number deduplication operation on a DataFrame.

See the docstring of the RowNumberDedup class for more information.
"""

from __future__ import annotations

from typing import List, Optional, Union

from pyspark.sql import Window, WindowSpec
from pyspark.sql.functions import col, desc, row_number

from koheesio.models import Field, conlist, field_validator
from koheesio.spark import Column
from koheesio.spark.transformations import ColumnsTransformation


class RowNumberDedup(ColumnsTransformation):
    """
    A class used to perform a row_number deduplication operation on a DataFrame.

    This class is a specialized transformation that extends the ColumnsTransformation class. It sorts the DataFrame
    based on the provided sort columns and assigns a row_number to each row. It then filters the DataFrame to keep only
    the top-row_number row for each group of duplicates.
    The row_number of each row can be stored in a specified target column or a default column named
    "meta_row_number_column". The class also provides an option to preserve meta columns
    (like the `row_number` column) in the output DataFrame.

    Attributes
    ----------
    columns : list
        List of columns to apply the transformation to. If a single '*' is passed as a column name or if the columns
        parameter is omitted, the transformation will be applied to all columns of the data types specified in
        `run_for_all_data_type` of the ColumnConfig. (inherited from ColumnsTransformation)
    sort_columns : list
        List of columns that the DataFrame will be sorted by.
    target_column : str, optional
        Column where the row_number of each row will be stored.
    preserve_meta : bool, optional
        Flag that determines whether the meta columns should be kept in the output DataFrame.
    """

    sort_columns: conlist(Union[str, Column], min_length=0) = Field(
        default_factory=list,
        alias="sort_column",
        description="List of orderBy columns. If only one column is passed, it can be passed as a single object.",
    )
    target_column: Optional[Union[str, Column]] = Field(
        default="meta_row_number_column",
        alias="target_suffix",
        description="The column to store the result in. If not provided, the result will be stored in the source"
        "column. Alias: target_suffix - if multiple columns are given as source, "
        "this will be used as a suffix",
    )
    preserve_meta: bool = Field(
        default=False,
        description="If true, meta columns are kept in output dataframe. Defaults to 'False'",
    )

    @field_validator("sort_columns", mode="before")
    def set_sort_columns(cls, columns_value: Union[str, Column, List[Union[str, Column]]]) -> List[Union[str, Column]]:
        """
        Validates and optimizes the sort_columns parameter.

        This method ensures that sort_columns is a list (or single object) of unique strings or Column objects.
        It removes any empty strings or None values from the list and deduplicates the columns.

        Parameters
        ----------
        columns_value : Union[str, Column, List[Union[str, Column]]]
            The value of the sort_columns parameter.

        Returns
        -------
        List[Union[str, Column]]
            The optimized and deduplicated list of sort columns.
        """
        columns = [columns_value] if isinstance(columns_value, (str, Column)) else [*columns_value]

        # Remove empty strings, None, etc.
        columns = [c for c in columns if (isinstance(c, Column) and c is not None) or (isinstance(c, str) and c)]

        dedup_columns = []
        seen = set()

        # Deduplicate the columns while preserving the order
        for column in columns:
            if str(column) not in seen:
                dedup_columns.append(column)
                seen.add(str(column))

        return dedup_columns

    @property
    def window_spec(self) -> WindowSpec:
        """
        Builds a WindowSpec object based on the columns defined in the configuration.

        The WindowSpec object is used to define a window frame over which functions are applied in Spark.
        This method partitions the data by the columns returned by the `get_columns` method and then orders the
        partitions by the columns specified in `sort_columns`.

        Notes
        -----
        The order of the columns in the WindowSpec object is preserved. If a column is passed as a string,
        it is converted to a Column object with DESC ordering.

        Returns
        -------
        WindowSpec
            A WindowSpec object that can be used to define a window frame in Spark.
        """
        # preserve the order of column object if it is passed, otherwise convert to column object
        # with DESC ordering
        order_clause = [desc(c) if isinstance(c, str) else c for c in self.sort_columns]

        return Window.partitionBy([*self.get_columns()]).orderBy(*order_clause)

    def execute(self) -> RowNumberDedup.Output:  # type: ignore
        """
        Performs the row_number deduplication operation on the DataFrame.

        This method sorts the DataFrame based on the provided sort columns, assigns a row_number to each row,
        and then filters the DataFrame to keep only the top-row_number row for each group of duplicates.
        The row_number of each row is stored in the target column. If preserve_meta is False,
        the method also drops the target column from the DataFrame.
        """
        df = self.df
        window_spec = self.window_spec

        # if target_column is a string, convert it to a Column object
        if isinstance((target_column := self.target_column), str):
            target_column = col(target_column)

        # dedup the dataframe based on the window spec
        df = df.withColumn(self.target_column, row_number().over(window_spec)).filter(target_column == 1).select("*")

        if not self.preserve_meta:
            df = df.drop(target_column)

        self.output.df = df



from typing import Any, List, Optional, Union
from abc import ABC, abstractmethod
import os
from pathlib import PurePath
from tempfile import TemporaryDirectory

from tableauhyperapi import (
    NOT_NULLABLE,
    NULLABLE,
    Connection,
    CreateMode,
    HyperProcess,
    Inserter,
    SqlType,
    TableDefinition,
    TableName,
    Telemetry,
)

from pydantic import Field, conlist

from pyspark.sql.functions import col
from pyspark.sql.types import (
    BooleanType,
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)

from koheesio.spark import DataFrame, SparkStep
from koheesio.spark.transformations.cast_to_datatype import CastToDatatype
from koheesio.spark.utils import SPARK_MINOR_VERSION
from koheesio.steps import Step, StepOutput


class HyperFile(Step, ABC):
    """
    Base class for all HyperFile classes
    """

    schema_: str = Field(default="Extract", alias="schema", description="Internal schema name within the Hyper file")
    table: str = Field(default="Extract", description="Table name within the Hyper file")

    @property
    def table_name(self) -> TableName:
        """
        Return TableName object for the Hyper file TableDefinition.
        """
        return TableName(self.schema_, self.table)


class HyperFileReader(HyperFile, SparkStep):
    """
    Read a Hyper file and return a Spark DataFrame.

    Examples
    --------
    ```python
    df = (
        HyperFileReader(
            path=PurePath(hw.hyper_path),
        )
        .execute()
        .df
    )
    ```
    """

    path: PurePath = Field(
        default=..., description="Path to the Hyper file", examples=["PurePath(~/data/my-file.hyper)"]
    )

    def execute(self) -> SparkStep.Output:
        type_mapping = {
            "date": StringType,
            "text": StringType,
            "double": FloatType,
            "bool": BooleanType,
            "small_int": ShortType,
            "big_int": LongType,
            "timestamp": StringType,
            "timestamp_tz": StringType,
            "int": IntegerType,
            "numeric": DecimalType,
        }
        df_cols = []
        timestamp_cols = []
        date_cols = []

        with HyperProcess(telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU) as hp:
            with Connection(endpoint=hp.endpoint, database=self.path) as connection:
                table_definition = connection.catalog.get_table_definition(name=self.table_name)

                select_cols = []
                self.log.debug(f"Schema for {self.table_name} in {self.path}:")
                for column in table_definition.columns:
                    self.log.debug(f"|-- {column.name}: {column.type} (nullable = {column.nullability})")

                    column_name = column.name.unescaped.__str__()
                    tableau_type = column.type.__str__().lower()

                    if tableau_type.startswith("numeric"):
                        spark_type = DecimalType(precision=18, scale=5)
                    else:
                        spark_type = type_mapping.get(tableau_type, StringType)()

                    if tableau_type == "timestamp" or tableau_type == "timestamp_tz":
                        timestamp_cols.append(column_name)
                        _col = f'cast("{column_name}" as text)'
                    elif tableau_type == "date":
                        date_cols.append(column_name)
                        _col = f'cast("{column_name}" as text)'
                    elif tableau_type.startswith("numeric"):
                        _col = f'cast("{column_name}" as decimal(18,5))'
                    else:
                        _col = f'"{column_name}"'

                    df_cols.append(StructField(column_name, spark_type))
                    select_cols.append(_col)

                data = connection.execute_list_query(f"select {','.join(select_cols)} from {self.table_name}")

        df_schema = StructType(df_cols)
        df = self.spark.createDataFrame(data, schema=df_schema)
        if timestamp_cols:
            df = CastToDatatype(column=timestamp_cols, datatype="timestamp").transform(df)
        if date_cols:
            df = CastToDatatype(column=date_cols, datatype="date").transform(df)

        self.output.df = df


class HyperFileWriter(HyperFile):
    """
    Base class for all HyperFileWriter classes

    Reference
    ---------
        HyperProcess parameters: https://tableau.github.io/hyper-db/docs/hyper-api/hyper_process/#process-settings
    """

    path: PurePath = Field(
        default=TemporaryDirectory().name,
        description="Path to the Hyper file, if executing in Databricks "
        "set the path manually and ensure to specify the scheme `dbfs:/`.",
        examples=["PurePath(/tmp/hyper/)", "PurePath(dbfs:/tmp/hyper/)"],
    )
    name: str = Field(default="extract", description="Name of the Hyper file")
    table_definition: TableDefinition = Field(
        default=None,
        description="Table definition to write to the Hyper file as described in "
        "https://tableau.github.io/hyper-db/lang_docs/py/tableauhyperapi.html#tableauhyperapi.TableDefinition",
    )
    hyper_process_parameters: dict = Field(
        # Disable logging by default, if logging is required remove the "log_config" key and refer to the Hyper API docs
        default={"log_config": ""},
        description="Set HyperProcess parameters, see Tableau Hyper API documentation for more details: "
        "https://tableau.github.io/hyper-db/docs/hyper-api/hyper_process/#process-settings",
    )

    class Output(StepOutput):
        """
        Output class for HyperFileListWriter
        """

        hyper_path: PurePath = Field(default=..., description="Path to created Hyper file")

    @property
    def hyper_path(self) -> PurePath:
        """
        Return full path to the Hyper file.
        """
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        hyper_path = PurePath(self.path, f"{self.name}.hyper" if ".hyper" not in self.name else self.name)
        self.log.info(f"Destination file: {hyper_path}")
        return hyper_path

    def write(self) -> Output:
        self.execute()

    @abstractmethod
    def execute(self) -> Output:
        pass


class HyperFileListWriter(HyperFileWriter):
    """
    Write list of rows to a Hyper file.

    Reference
    ---------
    Datatypes in https://tableau.github.io/hyper-db/docs/sql/datatype/ for supported data types.

    Examples
    --------
    ```python
    hw = HyperFileListWriter(
        name="test",
        table_definition=TableDefinition(
            table_name=TableName("Extract", "Extract"),
            columns=[
                TableDefinition.Column(
                    name="string",
                    type=SqlType.text(),
                    nullability=NOT_NULLABLE,
                ),
                TableDefinition.Column(
                    name="int", type=SqlType.int(), nullability=NULLABLE
                ),
                TableDefinition.Column(
                    name="timestamp",
                    type=SqlType.timestamp(),
                    nullability=NULLABLE,
                ),
            ],
        ),
        data=[
            ["text_1", 1, datetime(2024, 1, 1, 0, 0, 0, 0)],
            ["text_2", 2, datetime(2024, 1, 2, 0, 0, 0, 0)],
            ["text_3", None, None],
        ],
    ).execute()

    # do somthing with returned file path
    hw.hyper_path
    ```
    """

    data: conlist(List[Any], min_length=1) = Field(default=..., description="List of rows to write to the Hyper file")

    def execute(self):
        with HyperProcess(
            telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, parameters=self.hyper_process_parameters
        ) as hp:
            with Connection(
                endpoint=hp.endpoint, database=self.hyper_path, create_mode=CreateMode.CREATE_AND_REPLACE
            ) as connection:
                connection.catalog.create_schema(schema=self.table_definition.table_name.schema_name)
                connection.catalog.create_table(table_definition=self.table_definition)
                with Inserter(connection, self.table_definition) as inserter:
                    inserter.add_rows(rows=self.data)
                    inserter.execute()

        self.output.hyper_path = self.hyper_path


class HyperFileParquetWriter(HyperFileWriter):
    """
    Read one or multiple parquet files and write them to a Hyper file.

    Notes
    -----
    This method is much faster than HyperFileListWriter for large files.

    References
    ----------
    Copy from external format: https://tableau.github.io/hyper-db/docs/sql/command/copy_from
    Datatypes in https://tableau.github.io/hyper-db/docs/sql/datatype/ for supported data types.
    Parquet format limitations:
        https://tableau.github.io/hyper-db/docs/sql/external/formats/#external-format-parquet

    Examples
    --------
    ```python
    hw = HyperFileParquetWriter(
        name="test",
        table_definition=TableDefinition(
            table_name=TableName("Extract", "Extract"),
            columns=[
                TableDefinition.Column(
                    name="string",
                    type=SqlType.text(),
                    nullability=NOT_NULLABLE,
                ),
                TableDefinition.Column(
                    name="int", type=SqlType.int(), nullability=NULLABLE
                ),
                TableDefinition.Column(
                    name="timestamp",
                    type=SqlType.timestamp(),
                    nullability=NULLABLE,
                ),
            ],
        ),
        files=[
            "/my-path/parquet-1.snappy.parquet",
            "/my-path/parquet-2.snappy.parquet",
        ],
    ).execute()

    # do somthing with returned file path
    hw.hyper_path
    ```
    """

    file: conlist(Union[str, PurePath], min_length=1) = Field(
        default=..., alias="files", description="One or multiple parquet files to write to the Hyper file"
    )

    def execute(self) -> HyperFileWriter.Output:
        _file = [str(f) for f in self.file]
        array_files = "'" + "','".join(_file) + "'"

        with HyperProcess(
            telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU, parameters=self.hyper_process_parameters
        ) as hp:
            with Connection(
                endpoint=hp.endpoint, database=self.hyper_path, create_mode=CreateMode.CREATE_AND_REPLACE
            ) as connection:
                connection.catalog.create_schema(schema=self.table_definition.table_name.schema_name)
                connection.catalog.create_table(table_definition=self.table_definition)
                sql = f'copy "{self.schema_}"."{self.table}" from array [{array_files}] with (format parquet)'
                self.log.debug(f"Executing SQL: {sql}")
                connection.execute_command(sql)

        self.output.hyper_path = self.hyper_path


class HyperFileDataFrameWriter(HyperFileWriter):
    """
    Write a Spark DataFrame to a Hyper file.
    The process will write the DataFrame to a parquet file and then use the HyperFileParquetWriter to write to the
    Hyper file.

    Examples
    --------
    ```python
    hw = HyperFileDataFrameWriter(
        df=spark.createDataFrame(
            [(1, "foo"), (2, "bar")], ["id", "name"]
        ),
        name="test",
    ).execute()

    # or in Databricks
    hw = HyperFileDataFrameWriter(
        df=spark.createDataFrame(
            [(1, "foo"), (2, "bar")], ["id", "name"]
        ),
        name="test",
        path="dbfs:/tmp/hyper/",
    ).execute()

    # do somthing with returned file path
    hw.hyper_path
    ```
    """

    df: DataFrame = Field(default=..., description="Spark DataFrame to write to the Hyper file")
    table_definition: Optional[TableDefinition] = None  # table_definition is not required for this class

    @staticmethod
    def table_definition_column(column: StructField) -> TableDefinition.Column:
        """
        Convert a Spark StructField to a Tableau Hyper SqlType
        """
        type_mapping = {
            IntegerType(): SqlType.int,
            LongType(): SqlType.big_int,
            ShortType(): SqlType.small_int,
            DoubleType(): SqlType.double,
            FloatType(): SqlType.double,
            BooleanType(): SqlType.bool,
            DateType(): SqlType.date,
            StringType(): SqlType.text,
        }

        # Handling the TimestampNTZType for Spark 3.4+
        # Mapping both TimestampType and TimestampNTZType to NTZ type of Hyper
        if SPARK_MINOR_VERSION >= 3.4:
            from pyspark.sql.types import TimestampNTZType

            type_mapping[TimestampNTZType()] = SqlType.timestamp
            type_mapping[TimestampType()] = SqlType.timestamp
        # In older versions of Spark, only TimestampType is available and is mapped to TZ type of Hyper
        else:
            type_mapping[TimestampType()] = SqlType.timestamp_tz

        if column.dataType in type_mapping:
            sql_type = type_mapping[column.dataType]()  # type: ignore
        elif str(column.dataType).startswith("DecimalType"):
            # Tableau Hyper API limits the precision to 18 decimal places
            # noinspection PyUnresolvedReferences
            sql_type = SqlType.numeric(
                precision=column.dataType.precision if column.dataType.precision <= 18 else 18,
                scale=column.dataType.scale,
            )
        else:
            raise ValueError(f"Unsupported datatype '{column.dataType}' for column '{column.name}'.")

        return TableDefinition.Column(
            name=column.name, type=sql_type, nullability=NULLABLE if column.nullable else NOT_NULLABLE
        )

    @property
    def _table_definition(self) -> TableDefinition:
        schema = self.df.schema
        columns = list(map(self.table_definition_column, schema))

        td = TableDefinition(table_name=self.table_name, columns=columns)
        self.log.debug(f"Table definition for {self.table_name}:")
        for column in td.columns:
            self.log.debug(f"|-- {column.name}: {column.type} (nullable = {column.nullability})")

        return td

    def clean_dataframe(self) -> DataFrame:
        """
        - Replace NULLs for string and numeric columns
        - Convert data types to ensure compatibility with Tableau Hyper API
        """
        _df = self.df
        _schema = self.df.schema

        integer_cols = [field.name for field in _schema if field.dataType == IntegerType()]
        long_cols = [field.name for field in _schema if field.dataType == LongType()]
        short_cols = [field.name for field in _schema if field.dataType == ShortType()]
        double_cols = [field.name for field in _schema if field.dataType == DoubleType()]
        float_cols = [field.name for field in _schema if field.dataType == FloatType()]
        string_cols = [field.name for field in _schema if field.dataType == StringType()]
        timestamp_cols = [field.name for field in _schema if field.dataType == TimestampType()]

        # Handling the TimestampNTZType for Spark 3.4+
        # Any TimestampType column will be cast to TimestampNTZType for compatibility with Tableau Hyper API
        if SPARK_MINOR_VERSION >= 3.4:
            from pyspark.sql.types import TimestampNTZType

            for t_col in timestamp_cols:
                _df = _df.withColumn(t_col, col(t_col).cast(TimestampNTZType()))  # type: ignore

        # Replace null and NaN values with 0
        if len(integer_cols) > 0:
            _df = _df.na.fill(0, integer_cols)
        if len(long_cols) > 0:
            _df = _df.na.fill(0, long_cols)
        if len(short_cols) > 0:
            _df = _df.na.fill(0, short_cols)
        if len(double_cols) > 0:
            _df = _df.na.fill(0.0, double_cols)
        if len(float_cols) > 0:
            _df = _df.na.fill(0.0, float_cols)
        if len(string_cols) > 0:
            _df = _df.na.fill("", string_cols)

        # Cleanup decimal columns: enforce precision to 18, fill nulls with 0.0
        decimal_cols = [field for field in _schema if str(field.dataType).startswith("DecimalType")]
        decimal_col_names = []
        for d_col in decimal_cols:
            decimal_col_names.append(d_col.name)
            # noinspection PyUnresolvedReferences
            if d_col.dataType.precision > 18:
                # noinspection PyUnresolvedReferences
                _df = _df.withColumn(
                    d_col.name,
                    col(d_col.name).cast(DecimalType(precision=18, scale=d_col.dataType.scale)),  # type: ignore
                )
        if len(decimal_col_names) > 0:
            _df = _df.na.fill(0.0, decimal_col_names)

        return _df

    def write_parquet(self) -> List[PurePath]:
        _path = self.path.joinpath("parquet")
        (
            self.clean_dataframe()
            .coalesce(1)
            .write.option("delimiter", ",")
            .option("header", "true")
            .mode("overwrite")
            .parquet(_path.as_posix())
        )

        if _path.as_posix().startswith("dbfs:"):
            _path = PurePath(_path.as_posix().replace("dbfs:", "/dbfs"))
            self.log.debug("Parquet location on DBFS: %s}", _path)

        for _, _, files in os.walk(_path):
            for file in files:
                if file.endswith(".parquet"):
                    fp = PurePath(_path, file)
                    self.log.info("Parquet file created: %s", fp)
                    return [fp]

    def execute(self) -> HyperFileWriter.Output:
        w = HyperFileParquetWriter(
            path=self.path, name=self.name, table_definition=self._table_definition, files=self.write_parquet()
        )
        w.execute()
        self.output.hyper_path = w.output.hyper_path



from typing import Any, ContextManager, Optional, Union
from enum import Enum
import os
from pathlib import PurePath

from tableauserverclient import (
    DatasourceItem,
    PersonalAccessTokenAuth,
    ProjectItem,
    TableauAuth,
)
from tableauserverclient.server.pager import Pager
from tableauserverclient.server.server import Server
import urllib3  # type: ignore

from pydantic import Field, SecretStr

from koheesio.models import model_validator
from koheesio.steps import Step, StepOutput


class TableauServer(Step):
    """
    Base class for Tableau server interactions. Class provides authentication and project identification functionality.
    """

    url: str = Field(
        default=...,
        alias="url",
        description="Hostname for the Tableau server, e.g. tableau.my-org.com",
        examples=["tableau.my-org.com"],
    )
    user: str = Field(default=..., alias="user", description="Login name for the Tableau user")
    password: SecretStr = Field(default=..., alias="password", description="Password for the Tableau user")
    site_id: str = Field(
        default=...,
        alias="site_id",
        description="Identifier for the Tableau site, as used in the URL: https://tableau.my-org.com/#/site/SITE_ID",
    )
    version: str = Field(
        default="3.14",
        alias="version",
        description="Version of the Tableau server API",
    )
    token_name: Optional[str] = Field(
        default=None,
        alias="token_name",
        description="Name of the Tableau Personal Access Token",
    )
    token_value: Optional[SecretStr] = Field(
        default=None,
        alias="token_value",
        description="Value of the Tableau Personal Access Token",
    )
    project: Optional[str] = Field(
        default=None,
        alias="project",
        description="Name of the project on the Tableau server",
    )
    parent_project: Optional[str] = Field(
        default=None,
        alias="parent_project",
        description="Name of the parent project on the Tableau server, use 'root' for the root project.",
    )
    project_id: Optional[str] = Field(
        default=None,
        alias="project_id",
        description="ID of the project on the Tableau server",
    )

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.server: Optional[Server] = None

    @model_validator(mode="after")
    def validate_project(self) -> "TableauServer":
        """Validate when project and project_id are provided at the same time."""

        if self.project and self.project_id:
            raise ValueError("Both 'project' and 'project_id' parameters cannot be provided at the same time.")

        if not self.project and not self.project_id:
            raise ValueError("Either 'project' or 'project_id' parameters should be provided, none is set")

        return self

    @property
    def auth(self) -> ContextManager:
        """
        Authenticate on the Tableau server.

        Examples
        --------
        ```python
        with self._authenticate():
            self.server.projects.get()
        ```

        Returns
        -------
        ContextManager for TableauAuth or PersonalAccessTokenAuth authorization object
        """
        # Suppress 'InsecureRequestWarning'
        # noinspection PyUnresolvedReferences
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        tableau_auth: Union[TableauAuth, PersonalAccessTokenAuth]
        tableau_auth = TableauAuth(username=self.user, password=self.password.get_secret_value(), site_id=self.site_id)

        if self.token_name and self.token_value:
            self.log.info(
                "Token details provided, this will take precedence over username and password authentication."
            )
            tableau_auth = PersonalAccessTokenAuth(
                token_name=self.token_name, personal_access_token=self.token_value, site_id=self.site_id
            )

        self.server = Server(self.url)
        self.server.version = self.version
        self.server.add_http_options({"verify": False})

        return self.server.auth.sign_in(tableau_auth)

    @property
    def working_project(self) -> Union[ProjectItem, None]:
        """
        Identify working project by using `project` and `parent_project` (if necessary) class properties.
        The goal is to uniquely identify specific project on the server. If multiple projects have the same
        name, the `parent_project` attribute of the TableauServer is required.

        Notes
        -----
        Set `parent_project` value to 'root' if the project is located in the root directory.

        If `id` of the project is known, it can be used in `project_id` parameter, then the detection of the working
        project using the `project` and `parent_project` attributes is skipped.

        Returns
        -------
        ProjectItem object representing the working project
        """

        with self.auth:
            all_projects = Pager(self.server.projects)
            parent, lim_p = None, []

            for project in all_projects:
                if project.id == self.project_id:
                    lim_p = [project]
                    self.log.info(f"\nProject ID provided directly:\n\tName: {lim_p[0].name}\n\tID: {lim_p[0].id}")
                    break

                # Identify parent project
                if project.name.strip() == self.parent_project and not self.project_id:
                    parent = project
                    self.log.info(f"\nParent project identified:\n\tName: {parent.name}\n\tID: {parent.id}")

                # Identify project(s)
                if project.name.strip() == self.project and not self.project_id:
                    lim_p.append(project)

            # Further filter the list of projects by parent project id
            if self.parent_project == "root" and not self.project_id:
                lim_p = [p for p in lim_p if not p.parent_id]
            elif self.parent_project and parent and not self.project_id:
                lim_p = [p for p in lim_p if p.parent_id == parent.id]

            if len(lim_p) > 1:
                raise ValueError(
                    "Multiple projects with the same name exist on the server, "
                    "please provide `parent_project` attribute."
                )
            elif len(lim_p) == 0:
                raise ValueError("Working project could not be identified.")
            else:
                self.log.info(f"\nWorking project identified:\n\tName: {lim_p[0].name}\n\tID: {lim_p[0].id}")
                return lim_p[0]

    def execute(self) -> None:
        raise NotImplementedError("Method `execute` must be implemented in the subclass.")


class TableauHyperPublishMode(str, Enum):
    """
    Publishing modes for the TableauHyperPublisher.
    """

    APPEND = Server.PublishMode.Append
    OVERWRITE = Server.PublishMode.Overwrite


class TableauHyperPublisher(TableauServer):
    """
    Publish the given Hyper file to the Tableau server. Hyper file will be treated by Tableau server as a datasource.
    """

    datasource_name: str = Field(default=..., description="Name of the datasource to publish")
    hyper_path: PurePath = Field(default=..., description="Path to Hyper file")
    publish_mode: TableauHyperPublishMode = Field(
        default=TableauHyperPublishMode.OVERWRITE,
        description="Publish mode for the Hyper file",
    )

    class Output(StepOutput):
        """
        Output class for TableauHyperPublisher
        """

        datasource_item: DatasourceItem = Field(
            default=..., description="DatasourceItem object representing the published datasource"
        )

    def execute(self) -> None:
        # Ensure that the Hyper File exists
        if not os.path.isfile(self.hyper_path):
            raise FileNotFoundError(f"Hyper file not found at: {self.hyper_path.as_posix()}")

        with self.auth:
            # Finally, publish the Hyper File to the Tableau server
            self.log.info(f'Publishing Hyper File located at: "{self.hyper_path.as_posix()}"')
            self.log.debug(f"Create mode: {self.publish_mode}")

            datasource_item = self.server.datasources.publish(
                datasource_item=DatasourceItem(project_id=str(self.working_project.id), name=self.datasource_name),
                file=self.hyper_path.as_posix(),
                mode=self.publish_mode,
            )
            self.log.info(f"Published datasource to Tableau server with the id: {datasource_item.id}")

            self.output.datasource_item = datasource_item

    def publish(self) -> None:
        self.execute()



from textwrap import dedent

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.hash import Sha2Hash

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name="test_hash")


@pytest.mark.parametrize(
    "input_values,input_data,expected",
    [
        pytest.param(
            # input values,
            dict(target_column="hash", column="id"),
            None,
            # expected data
            [
                dict(id=1, string="hello", hash="6b86b273ff34fce19d6b804eff5a3f5747ada4eaa22f1d49c01e52ddb7875b4b"),
                dict(id=2, string="world", hash="d4735e3a265e16eee03f59718b9b5d03019c07d8b6c51f90da3a666eec13ab35"),
                dict(id=3, string="", hash="4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce"),
            ],
            id="base test",
        ),
        pytest.param(
            # input values,
            dict(target_column="hash", columns=["id", "string"]),
            None,
            # expected data
            [
                dict(id=1, string="hello", hash="79057cc0d6ced65f8dafc2cc0d7f93d628c3605f97e800416a344802dcf5407d"),
                dict(id=2, string="world", hash="939ac689cbc0affd630000575a8edb83583130ca570ab6000124c25f985603cf"),
                dict(id=3, string="", hash="626938c3fd46f4155cc77ec411626ef8188af25ad6829f2ad4c068ebc5d92fbf"),
            ],
            id="multiple column input",
        ),
        pytest.param(
            # input values,
            dict(target_column="hash", column="string"),
            ([["hello"], ["world"], [None]], ["string"]),
            # expected data
            [
                dict(string="hello", hash="2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824"),
                dict(string="world", hash="486ea46224d1bb4fb680f34f7c9ad96a8f24ec88be73ea8e5a6c65260e9cb8a7"),
                dict(string=None, hash=None),
            ],
            id="null column input",
        ),
        pytest.param(
            # input values,
            dict(target_column="hash", columns=["id", "string"]),
            ([[1, "hello"], [2, "world"], [3, None]], ["id", "string"]),
            # expected data
            [
                dict(id=1, string="hello", hash="79057cc0d6ced65f8dafc2cc0d7f93d628c3605f97e800416a344802dcf5407d"),
                dict(id=2, string="world", hash="939ac689cbc0affd630000575a8edb83583130ca570ab6000124c25f985603cf"),
                dict(id=3, string=None, hash="4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce"),
            ],
            id="null column input along with other columns",
        ),
    ],
)
def test_happy_flow(input_values, input_data, expected, sample_df_with_strings, spark):
    print(input_values)
    sha2_hash = Sha2Hash(**input_values)

    if input_data:
        input_df = spark.createDataFrame(*input_data)
    else:
        input_df = sample_df_with_strings

    df = sha2_hash.transform(input_df)
    actual = [k.asDict() for k in df.collect()]
    assert actual == expected
    log.info("HashUUID5 unit test completed")


def test_with_same_data_as_from_spark_docs(spark):
    input_df = spark.createDataFrame([["Alice"], ["Bob"]], ["name"])
    expected = [
        dict(name="Alice", sha2="3bc51062973c458d5a6f2d8d64a023246354ad7e064b1e4e009ec8a0699a3043"),
        dict(name="Bob", sha2="cd9fb1e148ccd8442e5aa74904cc73bf6fb54d1d54d333bd596aa9bb4bb4e961"),
    ]
    df = Sha2Hash(column="name", target_column="sha2", num_bits=256).transform(input_df)
    actual = [k.asDict() for k in df.collect()]
    assert actual == expected


@pytest.mark.parametrize(
    "input_values,error",
    [
        pytest.param(
            # input values,
            dict(target_column="hash", columns="column_does_not_exist"),
            # expected error
            ValueError,
            id="non-existent column",
        ),
    ],
)
def test_unhappy_flow(input_values, error, sample_df_with_strings):
    with pytest.raises(error):
        Sha2Hash(**input_values).transform(sample_df_with_strings)



"""
Excel reader for Spark

Note
----
Ensure the 'excel' extra is installed before using this reader.
Default implementation uses openpyxl as the engine for reading Excel files.
Other implementations can be used by passing the correct keyword arguments to the reader.

See Also
--------
- https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_excel.html
- koheesio.pandas.readers.excel.ExcelReader
"""

from pyspark.pandas import DataFrame as PandasDataFrame

from koheesio.pandas.readers.excel import ExcelReader as PandasExcelReader
from koheesio.spark.readers import Reader


class ExcelReader(Reader, PandasExcelReader):
    """Read data from an Excel file

    This class is a wrapper around the PandasExcelReader class. It reads an Excel file first using pandas, and then
    converts the pandas DataFrame to a Spark DataFrame.

    Attributes
    ----------
    path: str
        The path to the Excel file
    sheet_name: str
        The name of the sheet to read
    header: int
        The row to use as the column names
    """

    def execute(self) -> Reader.Output:
        pdf: PandasDataFrame = PandasExcelReader.from_step(self).execute().df
        self.output.df = self.spark.createDataFrame(pdf)



"""
Readers are a type of Step that read data from a source based on the input parameters and stores the result in
self.output.df.

For a comprehensive guide on the usage, examples, and additional features of Reader classes, please refer to the
[reference/spark/steps/readers](../../../reference/spark/readers.md) section of the Koheesio documentation.
"""

from abc import ABC, abstractmethod

from koheesio.models.reader import BaseReader
from koheesio.spark import SparkStep


class Reader(BaseReader, SparkStep, ABC):
    """Base class for all Readers

    A Reader is a Step that reads data from a source based on the input parameters
    and stores the result in self.output.df (DataFrame).

    When implementing a Reader, the execute() method should be implemented.
    The execute() method should read from the source and store the result in self.output.df.

    The Reader class implements a standard read() method that calls the execute() method and returns the result. This
    method can be used to read data from a Reader without having to call the execute() method directly. Read method
    does not need to be implemented in the child class.

    Every Reader has a SparkSession available as self.spark. This is the currently active SparkSession.

    The Reader class also implements a shorthand for accessing the output Dataframe through the df-property. If the
    output.df is None, .execute() will be run first.
    """

    @abstractmethod
    def execute(self) -> SparkStep.Output:
        """Execute on a Reader should handle self.output.df (output) as a minimum
        Read from whichever source -> store result in self.output.df
        """



import warnings

import pytest

from koheesio.spark.writers import BatchOutputMode

pytestmark = pytest.mark.spark


@pytest.mark.parametrize("output_mode", [(BatchOutputMode.APPEND)])
def test_muted_warnings(output_mode, dummy_df, spark):
    table_name = "test_table"

    with warnings.catch_warnings(record=True) as w:
        from koheesio.spark.writers.delta import DeltaTableWriter

        DeltaTableWriter(table=table_name, output_mode=output_mode, df=dummy_df)
        params_warnings = []

        for warning in w:
            # UserWarning: Field name "params" shadows an attribute in parent "DeltaTableWriter";
            if str(warning.message).startswith('Field name "params" shadows an attribute in parent'):
                print(str(warning.message))
                params_warnings.append(warning)

        assert len(params_warnings) == 0



"""Notification module for sending messages to notification services (e.g. Slack, Email, etc.)"""

from enum import Enum


class NotificationSeverity(Enum):
    """
    Enumeration of allowed message severities
    """

    INFO = "info"
    WARN = "warn"
    ERROR = "error"
    SUCCESS = "success"

    @property
    def alert_icon(self) -> str:
        """
        Return a colored circle in slack markup
        """
        return {
            NotificationSeverity.INFO: ":large_blue_circle:",
            NotificationSeverity.WARN: ":large_yellow_circle:",
            NotificationSeverity.ERROR: ":red_circle:",
            NotificationSeverity.SUCCESS: ":large_green_circle:",
        }.get(self, ":question:")



import datetime as dt

import pytest

from pyspark.sql import types as T

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.date_time.interval import (
    DateTimeAddInterval,
    DateTimeColumn,
    DateTimeSubtractInterval,
    adjust_time,
    col,
    dt_column,
    validate_interval,
)

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name="test_date_time")


@pytest.mark.parametrize(
    "input_data,column_name,operation,interval,expected",
    [
        (
            # "2019-01-01 00:00:00",
            dt.datetime(2019, 1, 1, 0, 0, 0),
            "datetime",
            "+",
            "1 day",
            dt.datetime(2019, 1, 2, 0, 0, 0),
        ),
        (
            "2019-01-01 00:00:00",
            "datetime",
            "-",
            "1 hour",
            "2018-12-31 23:00:00",
        ),
        (
            "2019-01-01",
            "date",
            "+",
            "1 day",
            "2019-01-02 00:00:00",
        ),
        (
            # 2019-01-01
            dt.datetime(2019, 1, 1, 0, 0, 0),
            # T.StructType([T.StructField("date", T.DateType())]),
            "date",
            "+",
            "10 months",
            # 2019-11-01
            dt.datetime(2019, 11, 1, 0, 0, 0),
        ),
        # (
        #     # Unhappy example !
        #     # ---
        #     # Seemingly there is a bug in Spark where an AttributeError is raised when trying to convert the use the
        #     # time column in the createDataFrame function. The error is not raised when using a date or datetime
        #     # column.
        #     # Error raised by pyspark: `AttributeError: 'datetime.time' object has no attribute 'timetuple'`
        #     # For this reason, we are leaving this test commented out for now...
        #     # ---
        #     # 00:00:00
        #     dt.time(0, 0, 0),
        #     "time",
        #     "+",
        #     "10 minutes",
        #     # 00:10:00
        #     dt.time(0, 10, 0),
        # ),
        (
            # try with microseconds
            # 2019-01-01 00:00:00.000
            dt.datetime(2019, 1, 1, 0, 0, 0, 0),
            "datetime",
            "+",
            "10 microseconds",
            # 2019-01-01 00:00:00.010
            dt.datetime(2019, 1, 1, 0, 0, 0, 10),
        ),
        (
            # check that we handle null values correctly
            None,
            "datetime",
            "+",
            "10 microseconds",
            None,
        ),
        (
            # complicated interval
            dt.datetime(2019, 1, 1, 0, 0, 0, 0),
            "datetime",
            "-",
            "5 days 3 hours 7 minutes 30 seconds 1 millisecond",
            dt.datetime(2018, 12, 26, 20, 52, 29, 999000),
        ),
    ],
)
def test_interval(input_data, column_name, operation, interval, expected, spark):
    if input_data is None:
        # create dataframe with null value and type string
        df = spark.createDataFrame([(input_data,)], T.StructType([T.StructField(column_name, T.StringType())]))
    else:
        df = spark.createDataFrame([(input_data,)], [column_name])

    column = col(column_name)
    column = DateTimeColumn.from_column(column)

    if operation == "-":
        df_adjusted = df.withColumn("adjusted", column - interval)
    elif operation == "+":
        df_adjusted = df.withColumn("adjusted", column + interval)
    else:
        raise RuntimeError(f"Invalid operation: {operation}")

    result = df_adjusted.collect()[0]["adjusted"]
    assert result == expected


def test_interval_unhappy(spark):
    with pytest.raises(ValueError):
        validate_interval("some random sym*bol*s")
    # invalid operation
    with pytest.raises(ValueError):
        _ = adjust_time(col("some_col"), "invalid operation", "1 day")
    # invalid interval
    with pytest.raises(ValueError):
        _ = adjust_time(col("some_col"), "add", "invalid value")
    # invalid input passed to the Transformation
    with pytest.raises(ValueError):
        DateTimeAddInterval(column="date", interval="1 foo")


def test_interval_column_transformations(df_with_all_types):
    df = df_with_all_types.select("timestamp")

    result_df = df
    add_interval = DateTimeAddInterval(column="timestamp", interval="1 month", target_column="date_plus_1_month")
    result_df = add_interval.transform(result_df)
    actual_add = result_df.head()["date_plus_1_month"]

    subtract_interval = DateTimeSubtractInterval(
        column="timestamp", interval="1 month", target_column="date_minus_1_month"
    )
    result_df = subtract_interval.transform(result_df)
    actual_subtract = result_df.head()["date_minus_1_month"]

    # expected data
    expected_add = dt.datetime(2023, 2, 1, 0, 1, 1)
    expected_subtract = dt.datetime(2022, 12, 1, 0, 1, 1)

    assert actual_add == expected_add
    assert actual_subtract == expected_subtract


def test_dt_column_with_string(spark):
    """Test that the dt_column function is a drop in replacement for pyspark.sql.functions.col given a string input"""
    df = spark.createDataFrame([(1, "2022-01-01 00:00:00")], ["id", "my_column"])
    result = df.select(dt_column("my_column")).collect()[0][0]
    assert result == "2022-01-01 00:00:00"


def test_dt_column_with_invalid_input():
    with pytest.raises(TypeError):
        dt_column(123)



from typing import Optional

from pyspark.sql import SparkSession

from koheesio.spark.utils.common import (
    check_if_pyspark_connect_is_supported,
    get_active_session,
)

__all__ = ["is_remote_session"]


def is_remote_session(spark: Optional[SparkSession] = None) -> bool:
    result = False

    if (_spark := spark or get_active_session()) and check_if_pyspark_connect_is_supported():
        # result = True if _spark.conf.get("spark.remote", None) else False  # type: ignore
        from pyspark.sql.connect.session import SparkSession as ConnectSparkSession

        result = isinstance(_spark, ConnectSparkSession)

    return result



from aiohttp import ClientSession, TCPConnector
from aiohttp_retry import ExponentialRetry
import pytest
import requests_mock
from yarl import URL

from pyspark.sql.types import MapType, StringType, StructField, StructType

from koheesio.asyncio.http import AsyncHttpStep
from koheesio.spark.readers.rest_api import AsyncHttpGetStep, RestApiReader
from koheesio.steps.http import PaginatedHttpGetStep

ASYNC_BASE_URL = "http://httpbin.org"
ASYNC_GET_ENDPOINT = URL(f"{ASYNC_BASE_URL}/get")

pytestmark = pytest.mark.spark


@pytest.fixture
def mock_paginated_api():
    with requests_mock.Mocker() as m:
        for i in range(1, 4):  # Mock 3 pages of data
            data = [{"id": j, "page": i, "value": f"data_{i}_{j}"} for j in range(1, 11)]  # 10 records per page
            m.get(f"https://api.example.com/data?page={i}", json=data)
        yield m


def test_paginated_api(mock_paginated_api):
    # Test that the paginated API returns all the data
    transport = PaginatedHttpGetStep(url="https://api.example.com/data?page={page}", paginate=True, pages=3)
    task = RestApiReader(transport=transport, spark_schema="id: int, page:int, value: string")

    assert isinstance(task.transport, PaginatedHttpGetStep)

    task.execute()

    # Convert the DataFrame to a list of dictionaries
    all_data = [row.asDict() for row in task.output.df.collect()]
    expected_data = [{"id": j, "page": i, "value": f"data_{i}_{j}"} for i in range(1, 4) for j in range(1, 11)]

    assert all_data == expected_data


@pytest.mark.asyncio
async def test_async_rest_api_reader():
    """
    Testing the AsyncHttpStep class.
    """
    session = ClientSession()
    urls = [URL(ASYNC_GET_ENDPOINT), URL(ASYNC_GET_ENDPOINT)]
    retry_options = ExponentialRetry()
    connector = TCPConnector(limit=10)
    headers = {"Content-Type": "application/json", "X-type": "Koheesio RestApiReader Test"}
    transport = AsyncHttpGetStep(
        client_session=session, url=urls, retry_options=retry_options, connector=connector, headers=headers
    )
    spark_schema = StructType(
        [
            StructField("args", MapType(StringType(), StringType()), True),
            StructField(
                "headers",
                StructType(
                    [
                        StructField("Accept", StringType(), True),
                        StructField("Accept-Encoding", StringType(), True),
                        StructField("Content-Type", StringType(), True),
                        StructField("Host", StringType(), True),
                        StructField("User-Agent", StringType(), True),
                        StructField("X-Amzn-Trace-Id", StringType(), True),
                        StructField("X-Type", StringType(), True),
                    ]
                ),
                True,
            ),
            StructField("origin", StringType(), True),
            StructField("url", StringType(), True),
        ]
    )
    task = RestApiReader(transport=transport, spark_schema=spark_schema)

    assert isinstance(task.transport, AsyncHttpStep)

    task.execute()

    rows = [row.asDict() for row in task.output.df.collect()]
    all_data = [{row["url"]: row["headers"].asDict()["X-Type"]} for row in rows]

    # Assert the responses_urls
    assert len(all_data) == 2
    assert all_data == [{f"{ASYNC_BASE_URL}/get": "Koheesio RestApiReader Test"}] * 2



"""
Module for reading data from JDBC sources.

Classes
-------
JdbcReader
    Reader for JDBC tables.
"""

from typing import Any, Dict, Optional

from koheesio import ExtraParamsMixin
from koheesio.models import Field, SecretStr, model_validator
from koheesio.spark.readers import Reader


class JdbcReader(Reader, ExtraParamsMixin):
    """
    Reader for JDBC tables.

    Wrapper around Spark's jdbc read format

    Notes
    -----
    * Query has precedence over dbtable. If query and dbtable both are filled in, dbtable will be ignored!
    * Extra options to the spark reader can be passed through the `options` input. Refer to Spark documentation
        for details: https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html
    * Consider using `fetchsize` as one of the options, as it is greatly increases the performance of the reader
    * Consider using `numPartitions`, `partitionColumn`, `lowerBound`, `upperBound` together with real or synthetic
        partitioning column as it will improve the reader performance

    When implementing a JDBC reader, the `get_options()` method should be implemented. The method should return a dict
    of options required for the specific JDBC driver. The `get_options()` method can be overridden in the child class.
    Additionally, the `driver` parameter should be set to the name of the JDBC driver. Be aware that the driver jar
    needs to be included in the Spark session; this class does not (and can not) take care of that!

    Example
    -------
    > Note: jars should be added to the Spark session manually. This class does not take care of that.

    This example depends on the jar for MS SQL:
     `https://repo1.maven.org/maven2/com/microsoft/sqlserver/mssql-jdbc/9.2.1.jre8/mssql-jdbc-9.2.1.jre8.jar`

    ```python
    from koheesio.spark.readers.jdbc import JdbcReader

    jdbc_mssql = JdbcReader(
        driver="com.microsoft.sqlserver.jdbc.SQLServerDriver",
        url="jdbc:sqlserver://10.xxx.xxx.xxx:1433;databaseName=YOUR_DATABASE",
        user="YOUR_USERNAME",
        password="***",
        dbtable="schema_name.table_name",
        options={
            "fetchsize": 100
        },  # you can also use 'params' instead of 'options'
    )
    df = jdbc_mssql.read()
    ```

    ### ExtraParamsMixin

    The `ExtraParamsMixin` is a mixin class that provides a way to pass extra parameters to the reader. The extra
    parameters are stored in the `params` (or `options`) attribute. Any key-value pairs passed to the reader will be
    stored in the `params` attribute.
    """

    format: str = Field(default="jdbc", description="The type of format to load. Defaults to 'jdbc'.")
    driver: str = Field(
        default=...,
        description="Driver name. Be aware that the driver jar needs to be passed to the task",
    )
    url: str = Field(
        default=...,
        description="URL for the JDBC driver. Note, in some environments you need to use the IP Address instead of "
        "the hostname of the server.",
    )
    user: str = Field(default=..., description="User to authenticate to the server")
    password: SecretStr = Field(default=..., description="Password belonging to the username")
    dbtable: Optional[str] = Field(
        default=None, description="Database table name, also include schema name", alias="table"
    )
    query: Optional[str] = Field(default=None, description="Query")
    params: Dict[str, Any] = Field(
        default_factory=dict, description="Extra options to pass to spark reader", alias="options"
    )

    def get_options(self) -> Dict[str, Any]:
        """
        Dictionary of options required for the specific JDBC driver.

        Note: override this method if driver requires custom names, e.g. Snowflake: `sfUrl`, `sfUser`, etc.
        """
        _options = {
            "driver": self.driver,
            "url": self.url,
            "user": self.user,
            "password": self.password,
            **self.params,
        }
        if query := self.query:
            _options["query"] = query

        # Check that only one of them is filled in
        if query and self.dbtable:
            self.log.warning("Query is filled in, dbtable will be ignored!")
        else:
            _options["dbtable"] = self.dbtable

        return _options

    @model_validator(mode="after")
    def check_dbtable_or_query(self) -> "JdbcReader":
        """Check that dbtable or query is filled in and that only one of them is filled in (query has precedence)"""
        # Check that dbtable or query is filled in
        if not self.dbtable and not self.query:
            raise ValueError("Please do not leave dbtable and query both empty!")

        return self

    @property
    def options(self) -> Dict[str, Any]:
        """Shorthand for accessing self.params provided for backwards compatibility"""
        return self.params

    def execute(self) -> "JdbcReader.Output":
        """Wrapper around Spark's jdbc read format"""
        options = self.get_options()

        if pw := self.password:
            options["password"] = pw.get_secret_value()

        if query := self.query:
            self.log.info(f"Executing query: {query}")

        self.output.df = self.spark.read.format(self.format).options(**options).load()



from typing import Any

import pytest
from tableauserverclient import DatasourceItem

from koheesio.integrations.spark.tableau.server import TableauHyperPublisher


class TestTableauServer:
    @pytest.fixture(autouse=False)
    def server(self, mocker):
        __server = mocker.patch("koheesio.integrations.spark.tableau.server.Server")
        __mock_server = __server.return_value

        from koheesio.integrations.spark.tableau.server import TableauServer

        # Mocking various returns from the Tableau server
        def create_mock_object(name_prefix: str, object_id: int, spec: Any = None, project_id: int = None):
            obj = mocker.MagicMock() if not spec else mocker.MagicMock(spec=spec)
            obj.id = f"{object_id}"
            obj.name = f"{name_prefix}-{object_id}"
            obj.project_id = f"{project_id}" if project_id else None
            return obj

        def create_mock_pagination(length: int):
            obj = mocker.MagicMock()
            obj.total_available = length
            return obj

        # Projects
        mock_projects = []
        mock_projects_pagination = mocker.MagicMock()

        def create_mock_project(name_prefix: str, project_id: int, parent_id: int = None):
            mock_project = mocker.MagicMock()
            mock_project.name = f"{name_prefix}-{project_id}"
            mock_project.id = f"{project_id}" if not parent_id else f"{project_id * parent_id}"
            mock_project.parent_id = f"{parent_id}" if parent_id else None
            return mock_project

        # Parent Projects
        for i in range(1, 3):
            mock_projects.append(create_mock_project(name_prefix="parent-project", project_id=i))
            # Child Projects
            r = range(3, 5) if i == 1 else range(4, 6)
            for ix in r:
                mock_projects.append(create_mock_project(name_prefix="project", project_id=ix, parent_id=i))

        mock_projects_pagination.total_available = len(mock_projects)

        __mock_server.projects.get.return_value = [
            mock_projects,
            mock_projects_pagination,
        ]

        # Data Sources
        mock_ds = create_mock_object("datasource", 1)
        mock_conn = create_mock_object("connection", 1)
        mock_conn.type = "baz"
        mock_ds.connections = [mock_conn]
        __mock_server.datasources.get.return_value = [
            [mock_ds],
            create_mock_pagination(1),
        ]

        __mock_server.datasources.publish.return_value = create_mock_object(
            "published_datasource", 1337, spec=DatasourceItem
        )

        yield TableauServer(
            url="https://tableau.domain.com", user="user", password="pass", site_id="site", project_id="1"
        )

    @pytest.fixture()
    def hyper_file(self, data_path):
        return f"{data_path}/readers/hyper_file/dummy.hyper"

    def test_working_project_w_project_id(self, server):
        server.project_id = "3"
        wp = server.working_project
        assert wp.id == "3" and wp.name == "project-3"

    def test_working_project_w_project_name(self, server):
        server.project_id = None
        server.project = "project-5"
        wp = server.working_project
        assert wp.id == "10" and wp.name == "project-5"

    def test_working_project_w_project_name_and_parent_project(self, server):
        server.project_id = None
        server.project = "project-4"
        server.parent_project = "parent-project-1"
        wp = server.working_project
        assert wp.id == "4" and wp.name == "project-4"

    def test_working_project_w_project_name_and_root(self, server):
        server.project_id = None
        server.project = "parent-project-1"
        server.parent_project = "root"
        wp = server.working_project
        assert wp.id == "1" and wp.name == "parent-project-1"

    def test_working_project_multiple_projects(self, server):
        with pytest.raises(ValueError):
            server.project_id = None
            server.project = "project-4"
            server.working_project

    def test_working_project_unknown(self, server):
        with pytest.raises(ValueError):
            server.project_id = None
            server.project = "project-6"
            server.working_project

    def test_publish_hyper(self, server, hyper_file):
        p = TableauHyperPublisher(
            url="https://tableau.domain.com",
            user="user",
            password="pass",
            site_id="site",
            project_id="1",
            hyper_path=hyper_file,
            datasource_name="published_datasource",
        )
        p.publish()
        assert p.output.datasource_item.id == "1337"



import pytest
import requests
from requests import HTTPError
from requests.adapters import HTTPAdapter
from requests.exceptions import RetryError
import requests_mock
from urllib3 import Retry

from koheesio.logger import LoggingFactory
from koheesio.models import SecretStr
from koheesio.steps.http import (
    HttpDeleteStep,
    HttpGetStep,
    HttpMethod,
    HttpPostStep,
    HttpPutStep,
    HttpStep,
)

BASE_URL = "https://httpbin.org"
GET_ENDPOINT = f"{BASE_URL}/get"
POST_ENDPOINT = f"{BASE_URL}/post"
PUT_ENDPOINT = f"{BASE_URL}/put"
DELETE_ENDPOINT = f"{BASE_URL}/delete"
STATUS_404_ENDPOINT = f"{BASE_URL}/status/404"
STATUS_500_ENDPOINT = f"{BASE_URL}/status/500"
STATUS_503_ENDPOINT = f"{BASE_URL}/status/503"


log = LoggingFactory.get_logger(name="test_delta", inherit_from_koheesio=True)


@pytest.mark.parametrize(
    "endpoint,step,method,return_value,expected_status_code",
    [
        pytest.param(
            GET_ENDPOINT,
            HttpGetStep,
            "get",
            {"job_id": "1234"},
            200,
            id="httpGetStep_success",
        ),
        pytest.param(GET_ENDPOINT, HttpGetStep, "get", {"message": "Error Code"}, 400, id="httpGetStep_error"),
        pytest.param(GET_ENDPOINT, HttpGetStep, "get", None, 404, id="httpGetStep_emptyresponse"),
        pytest.param(
            POST_ENDPOINT,
            HttpPostStep,
            "post",
            {"job_id": "1234"},
            200,
            id="httpPostStep_success",
        ),
        pytest.param(POST_ENDPOINT, HttpPostStep, "post", {"message": "Error Code"}, 400, id="httpPosStep_error"),
        pytest.param(
            PUT_ENDPOINT,
            HttpPutStep,
            "put",
            {"detail": "Updated"},
            202,
            id="httpPutStep_success",
        ),
        pytest.param(PUT_ENDPOINT, HttpPutStep, "put", {"message": "Error Code"}, 400, id="httpPutStep_error"),
        pytest.param(
            PUT_ENDPOINT, HttpPutStep, "put", {"message": "Validation Error"}, 422, id="httpPutStep_validationError"
        ),
        pytest.param(
            DELETE_ENDPOINT,
            HttpDeleteStep,
            "delete",
            {"detail": "Deleted"},
            204,
            id="httpDeleteStep_success",
        ),
        pytest.param(
            DELETE_ENDPOINT, HttpDeleteStep, "delete", {"message": "Error Code"}, 400, id="httpDeleteStep_error"
        ),
        pytest.param(
            DELETE_ENDPOINT,
            HttpDeleteStep,
            "delete",
            {"message": "Validation Error"},
            422,
            id="httpDeleteStep_validationError",
        ),
    ],
)
def test_http_step(endpoint: str, step: HttpStep, method: str, return_value: dict, expected_status_code: int) -> None:
    """
    Unit Testing the GET functions.
    Above parameters are for the success and failed GET API calls
    """

    with requests_mock.Mocker() as rm:
        rm.request(method=method, url=endpoint, json=return_value, status_code=int(expected_status_code))
        step = step(
            url=endpoint,
            headers={
                "Authorization": "Bearer token",
                "Content-Type": "application/json",
            },
        )
        # Unhappy path
        if expected_status_code not in (200, 202, 204):
            with pytest.raises(HTTPError) as excinfo:
                step.execute()

            assert excinfo.value.response.status_code == expected_status_code
        # Happy path
        else:
            step.execute()
            assert step.output.status_code == expected_status_code  # type: ignore
            assert step.output.response_json == return_value  # type: ignore


def test_http_step_with_valid_http_method():
    """
    Unit Testing the GET functions.
    Above parameters are for the success and failed GET API calls
    """

    with requests_mock.Mocker() as rm:
        rm.get(GET_ENDPOINT, status_code=int(200))
        response = HttpStep(method="get", url=GET_ENDPOINT).execute()
        assert response.status_code == 200


def test_http_step_with_invalid_http_method():
    with requests_mock.Mocker() as rm:
        rm.get(GET_ENDPOINT, status_code=int(200))
        # Will be raised during class instantiation
        with pytest.raises(AttributeError):
            HttpStep(method="foo", url=GET_ENDPOINT).execute()


def test_http_step_request():
    with requests_mock.Mocker() as rm:
        rm.put(PUT_ENDPOINT, status_code=int(200))
        # The default method for HttpStep class is GET, however the method specified in `request` options is PUT and
        # it will override the default
        step = HttpStep(url=PUT_ENDPOINT)
        with step._request(method=HttpMethod.PUT) as response:
            assert response.status_code == 200


EXAMPLE_DIGEST_AUTH = (
    'Digest username="Zaphod", realm="galaxy@heartofgold.com", nonce="42answer", uri="/dir/restaurant.html", '
    'response="dontpanic42", opaque="babelfish"'
)


@pytest.mark.parametrize(
    "params, expected",
    [
        pytest.param(
            dict(url=GET_ENDPOINT, headers={"Authorization": "Bearer token", "Content-Type": "application/json"}),
            "Bearer token",
            id="bearer_token_in_headers",
        ),
        pytest.param(
            dict(url=GET_ENDPOINT, headers={"Content-Type": "application/json"}, auth_header="Bearer token"),
            "Bearer token",
            id="bearer_token_through_auth_header",
        ),
        pytest.param(
            dict(url=GET_ENDPOINT, auth_header=EXAMPLE_DIGEST_AUTH),
            EXAMPLE_DIGEST_AUTH,
            id="digest_auth_with_nonce",
        ),
    ],
)
def test_get_headers(params: dict, expected: str, caplog: pytest.LogCaptureFixture) -> None:
    """
    Authorization headers are being converted into SecretStr under the hood to avoid dumping any
    sensitive content into logs. However, when calling the `get_headers` method, the SecretStr is being
    converted back to string, otherwise sensitive info would have looked like '**********'.
    """
    # Arrange and Act
    with requests_mock.Mocker() as rm:
        rm.get(params["url"], status_code=int(200))  # Mock the request to be always successful
        step = HttpStep(**params)
        caplog.set_level("DEBUG", logger=step.log.name)
        auth = step.headers.get("Authorization")
        step.execute()

    # Check that the token doesn't accidentally leak in the logs
    assert len(caplog.records) > 1, "No logs were generated"
    for record in caplog.records:
        assert expected not in record.message

    # Ensure that the Authorization header is properly parsed to a SecretStr
    assert auth is not None, "Authorization header is missing"
    assert isinstance(auth, SecretStr)
    assert str(auth) == "**********"
    assert auth.get_secret_value() == expected

    # Ensure that the Content-Type header is properly parsed while not being a SecretStr
    assert step.headers["Content-Type"] == "application/json"


@pytest.mark.parametrize(
    "max_retries, endpoint, status_code, expected_count, error_type",
    [
        pytest.param(4, STATUS_503_ENDPOINT, 503, 5, RetryError, id="max_retries_4_503"),
        pytest.param(7, STATUS_404_ENDPOINT, 404, 8, RetryError, id="max_retries_7_404"),
        pytest.param(17, STATUS_500_ENDPOINT, 500, 18, RetryError, id="max_retries_17_500"),
        pytest.param(17, STATUS_404_ENDPOINT, 503, 1, HTTPError, id="max_retries_17_404"),
    ],
)
def test_max_retries(
    max_retries: int, endpoint: str, status_code: int, expected_count: int, error_type: Exception
) -> None:
    session = requests.Session()
    retry_logic = Retry(total=max_retries, status_forcelist=[status_code])
    session.mount("https://", HTTPAdapter(max_retries=retry_logic))
    # noinspection HttpUrlsUsage
    session.mount("http://", HTTPAdapter(max_retries=retry_logic))

    step = HttpGetStep(url=endpoint, session=session)

    with pytest.raises(error_type):  # type: ignore
        step.execute()

    first_pool = [v for _, v in session.adapters["https://"].poolmanager.pools._container.items()][0]

    assert first_pool.num_requests == expected_count


@pytest.mark.parametrize(
    "backoff,expected",
    [
        pytest.param(7, [0, 14, 28], id="backoff_7"),
        pytest.param(3, [0, 6, 12], id="backoff_3"),
        pytest.param(2, [0, 4, 8], id="backoff_2"),
        pytest.param(1, [0, 2, 4], id="backoff_1"),
    ],
)
def test_initial_delay_and_backoff(monkeypatch: pytest.FixtureRequest, backoff: int, expected: list) -> None:
    session = requests.Session()
    retry_logic = Retry(total=3, backoff_factor=backoff, status_forcelist=[503])
    session.mount("https://", HTTPAdapter(max_retries=retry_logic))
    # noinspection HttpUrlsUsage
    session.mount("http://", HTTPAdapter(max_retries=retry_logic))

    step = HttpGetStep(
        url=STATUS_503_ENDPOINT,
        headers={
            "Authorization": "Bearer token",
            "Content-Type": "application/json",
        },
        session=session,
    )

    # Save the original get_backoff_time method
    original_get_backoff_time = Retry.get_backoff_time

    backoff_values = []

    # Define a new function to replace get_backoff_time
    def mock_get_backoff_time(self, *args, **kwargs):
        result = original_get_backoff_time(self, *args, **kwargs)
        backoff_values.append(result)
        return result

    # Use monkeypatch to replace get_backoff_time
    monkeypatch.setattr(Retry, "get_backoff_time", mock_get_backoff_time)

    with pytest.raises(RetryError):
        step.execute()

    # Restore the original get_backoff_time method
    monkeypatch.undo()

    assert backoff_values == expected



"""Read data from a Delta table and return a DataFrame or DataStream

Classes
-------
DeltaTableReader
    Reads data from a Delta table and returns a DataFrame
DeltaTableStreamReader
    Reads data from a Delta table and returns a DataStream
"""

from __future__ import annotations

from typing import Any, Dict, Optional, Union

from pydantic import PrivateAttr

from pyspark.sql import DataFrameReader
from pyspark.sql import functions as f

from koheesio.logger import LoggingFactory
from koheesio.models import Field, ListOfColumns, field_validator, model_validator
from koheesio.spark import Column, DataStreamReader
from koheesio.spark.delta import DeltaTableStep
from koheesio.spark.readers import Reader
from koheesio.utils import get_random_string

__all__ = ["DeltaTableReader", "DeltaTableStreamReader"]
STREAMING_SCHEMA_WARNING = (
    "\nImportant!\n"
    "Although you can start the streaming source from a specified version or timestamp, the schema of the streaming "
    "source is always the latest schema of the Delta table. You must ensure there is no incompatible schema change to "
    "the Delta table after the specified version or timestamp. Otherwise, the streaming source may return incorrect "
    "results when reading the data with an incorrect schema."
)

STREAMING_ONLY_OPTIONS = [
    "ignore_deletes",
    "ignore_changes",
    "starting_version",
    "starting_timestamp",
    "schema_tracking_location",
]


class DeltaTableReader(Reader):
    """Reads data from a Delta table and returns a DataFrame
    Delta Table can be read in batch or streaming mode
    It also supports reading change data feed (CDF) in both batch mode and streaming mode

    Parameters
    ----------
    table : Union[DeltaTableStep, str]
        The table to read
    filter_cond : Optional[Union[Column, str]]
        Filter condition to apply to the dataframe. Filters can be provided by using Column or string expressions.
        For example: `f.col('state') == 'Ohio'`, `state = 'Ohio'` or  `(col('col1') > 3) & (col('col2') < 9)`
    columns : Optional[ListOfColumns]
        Columns to select from the table. One or many columns can be provided as strings.
        For example: `['col1', 'col2']`, `['col1']` or `'col1'`
    streaming : Optional[bool]
        Whether to read the table as a Stream or not
    read_change_feed : bool
        readChangeFeed: Change Data Feed (CDF) feature allows Delta tables to track row-level changes between versions
        of a Delta table. When enabled on a Delta table, the runtime records 'change events' for all the data written
        into the table. This includes the row data along with metadata indicating whether the specified row was
        inserted, deleted, or updated. See: https://docs.databricks.com/delta/delta-change-data-feed.html
    starting_version : str
        startingVersion: The Delta Lake version to start from. All table changes starting from this version (inclusive)
        will be read by the streaming source. You can obtain the commit versions from the version column of the
        DESCRIBE HISTORY command output.
    starting_timestamp : str
        startingTimestamp: The timestamp to start from. All table changes committed at or after the timestamp
        (inclusive) will be read by the streaming source. Either provide a timestamp string
        (e.g. 2019-01-01T00:00:00.000Z) or a date string (e.g. 2019-01-01)
    ignore_deletes : bool
        ignoreDeletes: Ignore transactions that delete data at partition boundaries. Note: Only supported for
        streaming tables.
        For more info see https://docs.databricks.com/structured-streaming/delta-lake.html#ignore-updates-and-deletes
    ignore_changes : bool
        ignoreChanges: re-process updates if files had to be rewritten in the source table due to a data changing
        operation such as UPDATE, MERGE INTO, DELETE (within partitions), or OVERWRITE. Unchanged rows may still be
        emitted, therefore your downstream consumers should be able to handle duplicates. Deletes are not propagated
        downstream. ignoreChanges subsumes ignoreDeletes. Therefore, if you use ignoreChanges, your stream will not be
        disrupted by either deletions or updates to the source table.

    """

    table: Union[DeltaTableStep, str] = Field(default=..., description="The table to read")
    filter_cond: Optional[Union[Column, str]] = Field(
        default=None,
        alias="filterCondition",
        description="Filter condition to apply to the dataframe. Filters can be provided by using Column or string "
        "expressions For example: `f.col('state') == 'Ohio'`, `state = 'Ohio'` or  "
        "`(col('col1') > 3) & (col('col2') < 9)`",
    )
    columns: Optional[ListOfColumns] = Field(
        default=None,
        description="Columns to select from the table. One or many columns can be provided as strings. For example: "
        "`['col1', 'col2']`, `['col1']` or `'col1'` ",
    )

    # sql: Optional[str, Path]  # TODO: add support for SQL file, or advanced SQL expression
    streaming: Optional[bool] = Field(default=False, description="Whether to read the table as a Stream or not")
    read_change_feed: bool = Field(
        default=False,
        alias="readChangeFeed",
        description="Change Data Feed (CDF) feature allows Delta tables to track row-level changes between versions of "
        "a Delta table. When enabled on a Delta table, the runtime records 'change events' for all the data written "
        "into the table. This includes the row data along with metadata indicating whether the specified row was "
        "inserted, deleted, or updated. See: https://docs.databricks.com/delta/delta-change-data-feed.html",
    )

    # Specify initial position, one of:
    starting_version: Optional[str] = Field(
        default=None,
        alias="startingVersion",
        description="startingVersion: The Delta Lake version to start from. All table changes starting from this "
        "version (inclusive) will be read by the streaming source. You can obtain the commit versions from the version "
        "column of the DESCRIBE HISTORY command output." + STREAMING_SCHEMA_WARNING,
    )
    starting_timestamp: Optional[str] = Field(
        default=None,
        alias="startingTimestamp",
        description="startingTimestamp: The timestamp to start from. All table changes committed at or after the "
        "timestamp (inclusive) will be read by the streaming source. "
        "Either provide a timestamp string (e.g. 2019-01-01T00:00:00.000Z) or a date string (e.g. 2019-01-01)"
        + STREAMING_SCHEMA_WARNING,
    )

    # Streaming only options
    # ---
    # Ignore updates and deletes, one of:
    ignore_deletes: bool = Field(
        default=False,
        alias="ignoreDeletes",
        description="ignoreDeletes: Ignore transactions that delete data at partition boundaries. "
        "Note: Only supported for streaming tables. "
        "For more info see https://docs.databricks.com/structured-streaming/delta-lake.html#ignore-updates-and-deletes",
    )
    ignore_changes: bool = Field(
        default=False,
        alias="ignoreChanges",
        description="ignoreChanges: re-process updates if files had to be rewritten in the source table due to a data "
        "changing operation such as UPDATE, MERGE INTO, DELETE (within partitions), or OVERWRITE. Unchanged rows may "
        "still be emitted, therefore your downstream consumers should be able to handle duplicates. Deletes are not "
        "propagated downstream. ignoreChanges subsumes ignoreDeletes. Therefore if you use ignoreChanges, your stream "
        "will not be disrupted by either deletions or updates to the source table.",
    )
    skip_change_commits: bool = Field(
        default=False,
        alias="skipChangeCommits",
        description="skipChangeCommits: Skip processing of change commits. "
        "Note: Only supported for streaming tables. (not supported in Open Source Delta Implementation). "
        "Prefer using skipChangeCommits over ignoreDeletes and ignoreChanges starting DBR12.1 and above. "
        "For more info see https://docs.databricks.com/structured-streaming/delta-lake.html#skip-change-commits",
    )

    schema_tracking_location: Optional[str] = Field(
        default=None,
        alias="schemaTrackingLocation",
        description="schemaTrackingLocation: Track the location of source schema. "
        "Note: Recommend to enable Delta reader version: 3 and writer version: 7 for this option. "
        "For more info see https://docs.delta.io/latest/delta-column-mapping.html" + STREAMING_SCHEMA_WARNING,
    )

    # private attrs
    __temp_view_name__: Optional[str] = None
    __reader: Optional[Union[DataStreamReader, DataFrameReader]] = PrivateAttr(default=None)

    @property
    def temp_view_name(self) -> str:
        """Get the temporary view name for the dataframe for SQL queries"""
        return self.__temp_view_name__

    @field_validator("table")
    def _validate_table_name(cls, tbl: Union[DeltaTableStep, str]) -> DeltaTableStep:
        """Validate the table name provided as a string or a DeltaTableStep instance."""
        if isinstance(tbl, str):
            return DeltaTableStep(table=tbl)
        if isinstance(tbl, DeltaTableStep):
            return tbl
        raise AttributeError(f"Table name provided cannot be processed as a Table : {tbl}")

    @model_validator(mode="after")
    def _validate_starting_version_and_timestamp(self) -> "DeltaTableReader":
        """Validate 'starting_version' and 'starting_timestamp' - Only one of each should be provided"""
        starting_version = self.starting_version
        starting_timestamp = self.starting_timestamp
        streaming = self.streaming
        read_change_feed = self.read_change_feed

        if starting_version and starting_timestamp:
            raise ValueError(
                f"Specify either a 'starting_version' or a 'starting_timestamp', not both. \n"
                f"provided values: ({starting_version=}, {starting_timestamp=})"
            )

        if not streaming and read_change_feed and all([starting_version is None, starting_timestamp is None]):
            raise ValueError(
                "Specify either a 'starting_version' or a 'starting_timestamp' "
                "when reading change data feed in a batch mode."
            )

        return self

    @model_validator(mode="after")
    def _validate_ignore_deletes_and_changes_and_skip_commits(self) -> "DeltaTableReader":
        """Validate 'ignore_deletes' and 'ignore_changes' - Only one of each should be provided"""
        ignore_deletes = self.ignore_deletes
        ignore_changes = self.ignore_changes
        skip_change_commits = self.skip_change_commits

        if ignore_deletes and ignore_changes and skip_change_commits:
            raise ValueError(
                f"Specify either a 'ignore_deletes' or a 'ignore_changes', not both. \n"
                f"provided values: ({ignore_deletes=}, {ignore_changes=}, {skip_change_commits=})"
            )

        return self

    @model_validator(mode="before")
    def _warn_on_streaming_options_without_streaming(cls, options: Dict) -> Dict:
        """throws a warning if streaming options were provided, but streaming was not set to true"""
        streaming_options = [val for opt, val in options.items() if opt in STREAMING_ONLY_OPTIONS]
        streaming_toggled_on = options.get("streaming")

        if any(streaming_options) and not streaming_toggled_on:
            log = LoggingFactory.get_logger(name=cls.__name__, inherit_from_koheesio=True)
            log.warning(
                f"Streaming options were provided, but streaming was not toggled on. Was this intended?\n{options = }"
            )

        return options

    @model_validator(mode="after")
    def set_temp_view_name(self) -> "DeltaTableReader":
        """Set a temporary view name for the dataframe for SQL queries"""
        table_name = self.table.table
        vw_name = get_random_string(prefix=f"tmp_{table_name}")
        self.__temp_view_name__ = vw_name
        return self

    @property
    def view(self) -> str:
        """Create a temporary view of the dataframe for SQL queries"""
        temp_view_name = self.temp_view_name

        if (output_df := self.output.df) is None:
            self.log.warning(
                "Attempting to createTempView without any data being present. Please run .execute() or .read() first. "
                "View creation was unsuccessful."
            )
        else:
            output_df.createOrReplaceTempView(temp_view_name)

        return temp_view_name

    def get_options(self) -> Dict[str, Any]:
        """Get the options for the DeltaTableReader based on the `streaming` attribute"""
        options = {
            # Enable Change Data Feed (CDF) feature
            "readChangeFeed": self.read_change_feed,
            # Initial position, one of:
            "startingVersion": self.starting_version,
            "startingTimestamp": self.starting_timestamp,
        }

        # Streaming only options
        if self.streaming:
            options = {
                **options,
                # Ignore updates and deletes, one of:
                "ignoreDeletes": self.ignore_deletes,
                "ignoreChanges": self.ignore_changes,
                "skipChangeCommits": self.skip_change_commits,
                "schemaTrackingLocation": self.schema_tracking_location,
            }
        # Batch only options
        else:
            pass  # there are none... for now :)

        def normalize(v: Union[str, bool]) -> str:
            """normalize values"""
            # True becomes "true", False becomes "false"
            v = str(v).lower() if isinstance(v, bool) else v
            return v

        # Any options with `value == None` are filtered out
        return {k: normalize(v) for k, v in options.items() if v is not None}

    def __get_stream_reader(self) -> DataStreamReader:
        """Returns a basic DataStreamReader (streaming mode)"""
        return self.spark.readStream.format("delta")

    def __get_batch_reader(self) -> DataFrameReader:
        """Returns a basic DataFrameReader (batch mode)"""
        return self.spark.read.format("delta")

    @property
    def reader(self) -> Union[DataStreamReader, DataFrameReader]:
        """Return the reader for the DeltaTableReader based on the `streaming` attribute"""
        if not self.__reader:
            self.__reader = self.__get_stream_reader() if self.streaming else self.__get_batch_reader()
            self.__reader = self.__reader.options(**self.get_options())

        return self.__reader

    @reader.setter
    def reader(self, value: Union[DataStreamReader, DataFrameReader]):
        self.__reader = value

    def execute(self) -> Reader.Output:
        df = self.reader.table(self.table.table_name)
        if self.filter_cond is not None:
            df = df.filter(f.expr(self.filter_cond) if isinstance(self.filter_cond, str) else self.filter_cond)  # type: ignore
        if self.columns is not None:
            df = df.select(*self.columns)
        self.output.df = df

    # def read_from_sql_file(self):
    #     """TODO: implement create_dataframe_from_sql_file"""
    # TODO: I think it should method in DeltaTable, if you want to create table based on DDL
    # TODO: If you want to use SQL query for DataFrame, it make sense to add it to parent Reader class


class DeltaTableStreamReader(DeltaTableReader):
    """Reads data from a Delta table and returns a DataStream"""

    streaming: bool = True

    # def read_from_sql_file(self):
    #     """TODO: implement create_datastream_from_sql_file"""


# TODO: add support to extra read options for DeltaTableReader
